"""Tidy3d webapi types."""

from __future__ import annotations

import os
import pathlib
import tempfile
from datetime import datetime
from os import PathLike
from typing import Callable, Optional, Union

import requests
from botocore.exceptions import ClientError
from pydantic.v1 import Extra, Field, parse_obj_as

import tidy3d as td
from tidy3d.config import config
from tidy3d.exceptions import ValidationError

from . import http_util
from .cache import FOLDER_CACHE
from .constants import (
    SIM_ERROR_FILE,
    SIM_FILE_HDF5_GZ,
    SIM_LOG_FILE,
    SIM_VALIDATION_FILE,
    SIMULATION_DATA_HDF5_GZ,
)
from .core_config import get_logger_console
from .exceptions import WebError, WebNotFoundError
from .file_util import read_simulation_from_hdf5
from .http_util import get_version as _get_protocol_version
from .http_util import http
from .s3utils import download_file, download_gz_file, upload_file
from .stub import TaskStub
from .task_info import BatchDetail, TaskInfo
from .types import PayType, Queryable, ResourceLifecycle, Submittable, Tidy3DResource


class Folder(Tidy3DResource, Queryable, extra=Extra.allow):
    """Tidy3D Folder."""

    folder_id: str = Field(..., title="Folder id", description="folder id", alias="projectId")
    folder_name: str = Field(
        ..., title="Folder name", description="folder name", alias="projectName"
    )

    @classmethod
    def list(cls, projects_endpoint: str = "tidy3d/projects") -> []:
        """List all folders.

        Returns
        -------
        folders : [Folder]
            List of folders
        """
        resp = http.get(projects_endpoint)
        return (
            parse_obj_as(
                list[Folder],
                resp,
            )
            if resp
            else None
        )

    @classmethod
    def get(
        cls,
        folder_name: str,
        create: bool = False,
        projects_endpoint: str = "tidy3d/projects",
        project_endpoint: str = "tidy3d/project",
    ) -> Folder:
        """Get folder by name.

        Parameters
        ----------
        folder_name : str
            Name of the folder.
        create : str
            If the folder doesn't exist, create it.

        Returns
        -------
        folder : Folder
        """
        folder = FOLDER_CACHE.get(folder_name)
        if not folder:
            resp = http.get(project_endpoint, params={"projectName": folder_name})
            if resp:
                folder = Folder(**resp)
        if create and not folder:
            resp = http.post(projects_endpoint, {"projectName": folder_name})
            if resp:
                folder = Folder(**resp)
        FOLDER_CACHE[folder_name] = folder
        return folder

    @classmethod
    def create(cls, folder_name: str) -> Folder:
        """Create a folder, return existing folder if there is one has the same name.

        Parameters
        ----------
        folder_name : str
            Name of the folder.

        Returns
        -------
        folder : Folder
        """
        return Folder.get(folder_name, True)

    def delete(self, projects_endpoint: str = "tidy3d/projects") -> None:
        """Remove this folder."""

        http.delete(f"{projects_endpoint}/{self.folder_id}")

    def delete_old(self, days_old: int) -> int:
        """Remove folder contents older than ``days_old``."""

        return http.delete(
            f"tidy3d/tasks/{self.folder_id}/tasks",
            params={"daysOld": days_old},
        )

    def list_tasks(self, projects_endpoint: str = "tidy3d/projects") -> list[Tidy3DResource]:
        """List all tasks in this folder.

        Returns
        -------
        tasks : List[:class:`.SimulationTask`]
            List of tasks in this folder
        """
        resp = http.get(f"{projects_endpoint}/{self.folder_id}/tasks")
        return (
            parse_obj_as(
                list[SimulationTask],
                resp,
            )
            if resp
            else None
        )


class WebTask(ResourceLifecycle, Submittable, extra=Extra.allow):
    """Interface for managing the running a task on the server."""

    task_id: Optional[str] = Field(
        ...,
        title="task_id",
        description="Task ID number, set when the task is uploaded, leave as None.",
        alias="taskId",
    )

    @classmethod
    def create(
        cls,
        task_type: str,
        task_name: str,
        folder_name: str = "default",
        callback_url: Optional[str] = None,
        simulation_type: str = "tidy3d",
        parent_tasks: Optional[list[str]] = None,
        file_type: str = "Gz",
        projects_endpoint: str = "tidy3d/projects",
    ) -> SimulationTask:
        """Create a new task on the server.

        Parameters
        ----------
        task_type: :class".TaskType"
            The type of task.
        task_name: str
            The name of the task.
        folder_name: str,
            The name of the folder to store the task. Default is "default".
        callback_url: str
            Http PUT url to receive simulation finish event. The body content is a json file with
            fields ``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.
        simulation_type : str
            Type of simulation being uploaded.
        parent_tasks : List[str]
            List of related task ids.
        file_type: str
            the simulation file type Json, Hdf5, Gz

        Returns
        -------
        :class:`SimulationTask`
            :class:`SimulationTask` object containing info about status, size,
            credits of task and others.
        """

        # handle backwards compatibility, "tidy3d" is the default simulation_type
        if simulation_type is None:
            simulation_type = "tidy3d"

        folder = Folder.get(folder_name, create=True)

        if task_type in ["RF", "TERMINAL_CM", "MODAL_CM"]:
            payload = {
                "groupName": task_name,
                "folderId": folder.folder_id,
                "fileType": file_type,
                "taskType": task_type,
            }
            resp = http.post("rf/task", payload)
        else:
            payload = {
                "taskName": task_name,
                "taskType": task_type,
                "callbackUrl": callback_url,
                "simulationType": simulation_type,
                "parentTasks": parent_tasks,
                "fileType": file_type,
            }
            resp = http.post(f"{projects_endpoint}/{folder.folder_id}/tasks", payload)

        return SimulationTask(**resp, taskType=task_type, folder_name=folder_name)

    def get_url(self) -> str:
        base = str(config.web.website_endpoint or "")
        if isinstance(self, BatchTask):
            return "/".join([base.rstrip("/"), f"rf?taskId={self.task_id}"])
        return "/".join([base.rstrip("/"), f"workbench?taskId={self.task_id}"])

    def get_folder_url(self) -> Optional[str]:
        folder_id = getattr(self, "folder_id", None)
        if not folder_id:
            return None
        base = str(config.web.website_endpoint or "")
        return "/".join([base.rstrip("/"), f"folders/{folder_id}"])

    def get_log(
        self,
        to_file: PathLike,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> pathlib.Path:
        """Get log file from Server.

        Parameters
        ----------
        to_file: PathLike
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """

        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        target_path = pathlib.Path(to_file)

        return download_file(
            self.task_id,
            SIM_LOG_FILE,
            to_file=target_path,
            verbose=verbose,
            progress_callback=progress_callback,
        )

    def get_data_hdf5(
        self,
        to_file: PathLike,
        remote_data_file_gz: PathLike = SIMULATION_DATA_HDF5_GZ,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> pathlib.Path:
        """Download data artifact (simulation or batch) with gz fallback handling.

        Parameters
        ----------
        remote_data_file_gz : PathLike
            Gzipped remote filename.
        to_file : PathLike
            Local target path.
        verbose : bool
            Whether to log progress.
        progress_callback : Optional[Callable[[float], None]]
            Progress callback.

        Returns
        -------
        pathlib.Path
            Saved local path.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")
        target_path = pathlib.Path(to_file)
        file = None
        try:
            file = download_gz_file(
                resource_id=self.task_id,
                remote_filename=remote_data_file_gz,
                to_file=target_path,
                verbose=verbose,
                progress_callback=progress_callback,
            )
        except ClientError:
            if verbose:
                console = get_logger_console()
                console.log(f"Unable to download '{remote_data_file_gz}'.")
        if not file:
            try:
                file = download_file(
                    resource_id=self.task_id,
                    remote_filename=str(remote_data_file_gz)[:-3],
                    to_file=target_path,
                    verbose=verbose,
                    progress_callback=progress_callback,
                )
            except Exception as e:
                raise WebError(
                    "Failed to download the data file from the server. "
                    "Please confirm that the task completed successfully."
                ) from e
        return file

    @staticmethod
    def is_batch(resource_id: str) -> bool:
        """Checks if a given resource ID corresponds to a valid batch task.

        This is a utility function to verify a batch task's existence before
        instantiating the class.

        Parameters
        ----------
        resource_id : str
            The unique identifier for the resource.

        Returns
        -------
        bool
            ``True`` if the resource is a valid batch task, ``False`` otherwise.
        """
        try:
            # TODO PROPERLY FIXME
            # Disable non critical logs due to check for resourceId, until we have a dedicated API for this
            resp = http.get(
                f"rf/task/{resource_id}/statistics",
                suppress_404=True,
            )
            status = bool(resp and isinstance(resp, dict) and "status" in resp)
            return status
        except Exception:
            return False

    def delete(self, versions: bool = False) -> None:
        """Delete current task from server.

        Parameters
        ----------
        versions : bool = False
            If ``True``, delete all versions of the task in the task group. Otherwise, delete only
            the version associated with the current task ID.
        """
        if not self.task_id:
            raise ValueError("Task id not found.")

        task_details = self.detail().dict()

        if task_details and "groupId" in task_details:
            group_id = task_details["groupId"]
            if versions:
                http.delete("tidy3d/group", json={"groupIds": [group_id]})
                return
            elif "version" in task_details:
                version = task_details["version"]
                http.delete(f"tidy3d/group/{group_id}/versions", json={"versions": [version]})
                return

        # Fallback to old method if we can't get the groupId and version
        http.delete(f"tidy3d/tasks/{self.task_id}")


class SimulationTask(WebTask):
    """Interface for managing the running of solver tasks on the server."""

    folder_id: Optional[str] = Field(
        None,
        title="folder_id",
        description="Folder ID number, set when the task is uploaded, leave as None.",
        alias="folderId",
    )
    status: Optional[str] = Field(title="status", description="Simulation task status.")

    real_flex_unit: float = Field(
        None, title="real FlexCredits", description="Billed FlexCredits.", alias="realCost"
    )

    created_at: Optional[datetime] = Field(
        title="created_at", description="Time at which this task was created.", alias="createdAt"
    )

    task_type: Optional[str] = Field(
        title="task_type", description="The type of task.", alias="taskType"
    )

    folder_name: Optional[str] = Field(
        "default",
        title="Folder Name",
        description="Name of the folder associated with this task.",
        alias="folderName",
    )

    callback_url: str = Field(
        None,
        title="Callback URL",
        description="Http PUT url to receive simulation finish event. "
        "The body content is a json file with fields "
        "``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.",
    )

    # simulation_type: str = pd.Field(
    #     None,
    #     title="Simulation Type",
    #     description="Type of simulation, used internally only.",
    # )

    # parent_tasks: Tuple[TaskId, ...] = pd.Field(
    #     None,
    #     title="Parent Tasks",
    #     description="List of parent task ids for the simulation, used internally only."
    # )

    @classmethod
    def get(cls, task_id: str, verbose: bool = True) -> SimulationTask:
        """Get task from the server by id.

        Parameters
        ----------
        task_id: str
            Unique identifier of task on server.
        verbose:
            If `True`, will print progressbars and status, otherwise, will run silently.

        Returns
        -------
        :class:`.SimulationTask`
            :class:`.SimulationTask` object containing info about status,
             size, credits of task and others.
        """
        try:
            resp = http.get(f"tidy3d/tasks/{task_id}/detail")
        except WebNotFoundError as e:
            td.log.error(f"The requested task ID '{task_id}' does not exist.")
            raise e

        task = SimulationTask(**resp) if resp else None
        return task

    @classmethod
    def get_running_tasks(cls) -> list[SimulationTask]:
        """Get a list of running tasks from the server"

        Returns
        -------
        List[:class:`.SimulationTask`]
            :class:`.SimulationTask` object containing info about status,
             size, credits of task and others.
        """
        resp = http.get("tidy3d/py/tasks")
        if not resp:
            return []
        return parse_obj_as(list[SimulationTask], resp)

    def detail(self) -> TaskInfo:
        """Fetches the detailed information and status of the task.

        Returns
        -------
        TaskInfo
            An object containing the task's latest data.
        """
        resp = http.get(f"tidy3d/tasks/{self.task_id}/detail")
        return TaskInfo(**{"taskId": self.task_id, "taskType": self.task_type, **resp})

    def get_simulation_json(self, to_file: PathLike, verbose: bool = True) -> None:
        """Get json file for a :class:`.Simulation` from server.

        Parameters
        ----------
        to_file: PathLike
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        to_file = pathlib.Path(to_file)

        hdf5_file, hdf5_file_path = tempfile.mkstemp(".hdf5")
        os.close(hdf5_file)
        try:
            self.get_simulation_hdf5(hdf5_file_path)
            if os.path.exists(hdf5_file_path):
                json_string = read_simulation_from_hdf5(hdf5_file_path)
                to_file.parent.mkdir(parents=True, exist_ok=True)
                with to_file.open("w", encoding="utf-8") as file:
                    # Write the string to the file
                    file.write(json_string.decode("utf-8"))
                    if verbose:
                        console = get_logger_console()
                        console.log(f"Generate {to_file} successfully.")
            else:
                raise WebError("Failed to download simulation.json.")
        finally:
            os.unlink(hdf5_file_path)

    def upload_simulation(
        self,
        stub: TaskStub,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
        remote_sim_file: PathLike = SIM_FILE_HDF5_GZ,
    ) -> None:
        """Upload :class:`.Simulation` object to Server.

        Parameters
        ----------
        stub: :class:`TaskStub`
            An instance of TaskStub.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while uploading the data.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")
        if not stub:
            raise WebError("Expected field 'simulation' is unset.")
        # Also upload hdf5.gz containing all data.
        file, file_name = tempfile.mkstemp()
        os.close(file)
        try:
            # upload simulation
            # compress .hdf5 to .hdf5.gz
            stub.to_hdf5_gz(file_name)
            upload_file(
                self.task_id,
                file_name,
                remote_sim_file,
                verbose=verbose,
                progress_callback=progress_callback,
            )
        finally:
            os.unlink(file_name)

    def upload_file(
        self,
        local_file: PathLike,
        remote_filename: str,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> None:
        """
        Upload file to platform. Using this method when the json file is too large to parse
         as :class".simulation".
        Parameters
        ----------
        local_file: PathLike
            Local file path.
        remote_filename: str
            file name on the server
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while uploading the data.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        upload_file(
            self.task_id,
            local_file,
            remote_filename,
            verbose=verbose,
            progress_callback=progress_callback,
        )

    def submit(
        self,
        solver_version: Optional[str] = None,
        worker_group: Optional[str] = None,
        pay_type: Union[PayType, str] = PayType.AUTO,
        priority: Optional[int] = None,
    ) -> None:
        """Kick off this task.

        It will be uploaded to server before
        starting the task. Otherwise, this method assumes that the Simulation has been uploaded by
        the upload_file function, so the task will be kicked off directly.

        Parameters
        ----------
        solver_version: str = None
            target solver version.
        worker_group: str = None
            worker group
        pay_type: Union[PayType, str] = PayType.AUTO
            Which method to pay the simulation.
        priority: int = None
            Priority of the simulation in the Virtual GPU (vGPU) queue (1 = lowest, 10 = highest).
            It affects only simulations from vGPU licenses and does not impact simulations using FlexCredits.
        """
        pay_type = PayType(pay_type) if not isinstance(pay_type, PayType) else pay_type

        if solver_version:
            protocol_version = None
        else:
            protocol_version = http_util.get_version()

        http.post(
            f"tidy3d/tasks/{self.task_id}/submit",
            {
                "solverVersion": solver_version,
                "workerGroup": worker_group,
                "protocolVersion": protocol_version,
                "enableCaching": config.web.enable_caching,
                "payType": pay_type.value,
                "priority": priority,
            },
        )

    def estimate_cost(self, solver_version: Optional[str] = None) -> float:
        """Compute the maximum flex unit charge for a given task, assuming the simulation runs for
        the full ``run_time``. If early shut-off is triggered, the cost is adjusted proportionately.

        Parameters
        ----------
        solver_version: str
            target solver version.

        Returns
        -------
        flex_unit_cost: float
            estimated cost in FlexCredits
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        if solver_version:
            protocol_version = None
        else:
            protocol_version = http_util.get_version()

        resp = http.post(
            f"tidy3d/tasks/{self.task_id}/metadata",
            {
                "solverVersion": solver_version,
                "protocolVersion": protocol_version,
            },
        )
        return resp

    def get_simulation_hdf5(
        self,
        to_file: PathLike,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
        remote_sim_file: PathLike = SIM_FILE_HDF5_GZ,
    ) -> pathlib.Path:
        """Get simulation.hdf5 file from Server.

        Parameters
        ----------
        to_file: PathLike
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        target_path = pathlib.Path(to_file)

        return download_gz_file(
            resource_id=self.task_id,
            remote_filename=remote_sim_file,
            to_file=target_path,
            verbose=verbose,
            progress_callback=progress_callback,
        )

    def get_running_info(self) -> tuple[float, float]:
        """Gets the % done and field_decay for a running task.

        Returns
        -------
        perc_done : float
            Percentage of run done (in terms of max number of time steps).
            Is ``None`` if run info not available.
        field_decay : float
            Average field intensity normalized to max value (1.0).
            Is ``None`` if run info not available.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        resp = http.get(f"tidy3d/tasks/{self.task_id}/progress")
        perc_done = resp.get("perc_done")
        field_decay = resp.get("field_decay")
        return perc_done, field_decay

    def get_log(
        self,
        to_file: PathLike,
        verbose: bool = True,
        progress_callback: Optional[Callable[[float], None]] = None,
    ) -> pathlib.Path:
        """Get log file from Server.

        Parameters
        ----------
        to_file: PathLike
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.
        progress_callback : Callable[[float], None] = None
            Optional callback function called while downloading the data.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """

        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        target_path = pathlib.Path(to_file)

        return download_file(
            self.task_id,
            SIM_LOG_FILE,
            to_file=target_path,
            verbose=verbose,
            progress_callback=progress_callback,
        )

    def get_error_json(
        self, to_file: PathLike, verbose: bool = True, validation: bool = False
    ) -> pathlib.Path:
        """Get error json file for a :class:`.Simulation` from server.

        Parameters
        ----------
        to_file: PathLike
            Save file to path.
        verbose: bool = True
            Whether to display progress bars.
        validation: bool = False
            Whether to get a validation error file or a solver error file.

        Returns
        -------
        path: pathlib.Path
            Path to saved file.
        """
        if not self.task_id:
            raise WebError("Expected field 'task_id' is unset.")

        target_path = pathlib.Path(to_file)
        target_file = SIM_ERROR_FILE if not validation else SIM_VALIDATION_FILE

        return download_file(
            self.task_id,
            target_file,
            to_file=target_path,
            verbose=verbose,
        )

    def abort(self) -> requests.Response:
        """Abort the current task on the server."""
        if not self.task_id:
            raise ValueError("Task id not found.")
        return http.put(
            "tidy3d/tasks/abort", json={"taskType": self.task_type, "taskId": self.task_id}
        )

    def validate_post_upload(self, parent_tasks: Optional[list[str]] = None) -> None:
        """Perform checks after task is uploaded and metadata is processed."""
        if self.task_type == "HEAT_CHARGE" and parent_tasks:
            try:
                if len(parent_tasks) > 1:
                    raise ValueError(
                        "A single parent 'task_id' corresponding to the task in which the meshing "
                        "was run must be provided."
                    )
                try:
                    # get mesh task info
                    mesh_task = SimulationTask.get(parent_tasks[0], verbose=False)
                    assert mesh_task.task_type == "VOLUME_MESH"
                    assert mesh_task.status == "success"
                    # get up-to-date task info
                    task = SimulationTask.get(self.task_id, verbose=False)
                    if task.fileMd5 != mesh_task.childFileMd5:
                        raise ValidationError(
                            "Simulation stored in parent task 'VolumeMesher' does not match the "
                            "current simulation."
                        )
                except Exception as e:
                    raise ValidationError(
                        "The parent task must be a 'VolumeMesher' task which has been successfully "
                        "run and is associated to the same 'HeatChargeSimulation' as provided here."
                    ) from e

            except Exception as e:
                raise WebError(f"Provided 'parent_tasks' failed validation: {e!s}") from e


class BatchTask(WebTask):
    """Interface for managing a batch task on the server."""

    task_type: Optional[str] = Field(
        None, title="task_type", description="The type of task.", alias="taskType"
    )

    @classmethod
    def get(cls, task_id: str, verbose: bool = True) -> BatchTask:
        """Get batch task by id.

        Parameters
        ----------
        task_id: str
            Unique identifier of batch on server.
        verbose:
            If `True`, will print progressbars and status, otherwise, will run silently.

        Returns
        -------
        :class:`.BatchTask` | None
            BatchTask object if found, otherwise None.
        """
        try:
            resp = http.get(f"rf/task/{task_id}/statistics")
        except WebNotFoundError as e:
            td.log.error(f"The requested batch ID '{task_id}' does not exist.")
            raise e
        # Extract taskType from response if available
        if resp:
            task_type = resp.get("taskType") if isinstance(resp, dict) else None
            return BatchTask(taskId=task_id, taskType=task_type)
        return None

    def detail(self) -> BatchDetail:
        """Fetches the detailed information and status of the batch.

        Returns
        -------
        BatchDetail
            An object containing the batch's latest data.
        """
        resp = http.get(
            f"rf/task/{self.task_id}/statistics",
        )
        # Some backends may return null for collection fields; coerce to sensible defaults
        if isinstance(resp, dict):
            if resp.get("tasks") is None:
                resp["tasks"] = []
        return BatchDetail(**(resp or {}))

    def check(
        self,
        check_task_type: str,
        solver_version: Optional[str] = None,
        protocol_version: Optional[str] = None,
    ) -> requests.Response:
        """Submits a request to validate the batch configuration on the server.

        Parameters
        ----------
        solver_version : Optional[str], default=None
            The version of the solver to use for validation.
        protocol_version : Optional[str], default=None
            The data protocol version. Defaults to the current version.

        Returns
        -------
        Any
            The server's response to the check request.
        """
        if protocol_version is None:
            protocol_version = _get_protocol_version()
        return http.post(
            f"rf/task/{self.task_id}/check",
            {
                "solverVersion": solver_version,
                "protocolVersion": protocol_version,
                "taskType": check_task_type,
            },
        )

    def submit(
        self,
        solver_version: Optional[str] = None,
        protocol_version: Optional[str] = None,
        worker_group: Optional[str] = None,
        pay_type: Union[PayType, str] = PayType.AUTO,
        priority: Optional[int] = None,
    ) -> requests.Response:
        """Submits the batch for execution on the server.

        Parameters
        ----------
        solver_version : Optional[str], default=None
            The version of the solver to use for execution.
        protocol_version : Optional[str], default=None
            The data protocol version. Defaults to the current version.
        worker_group : Optional[str], default=None
            Optional identifier for a specific worker group to run on.

        Returns
        -------
        Any
            The server's response to the submit request.
        """

        # TODO: add support for pay_type and priority arguments
        if pay_type != PayType.AUTO:
            raise NotImplementedError(
                "The 'pay_type' argument is not yet supported and will be ignored."
            )
        if priority is not None:
            raise NotImplementedError(
                "The 'priority' argument is not yet supported and will be ignored."
            )

        if protocol_version is None:
            protocol_version = _get_protocol_version()
        return http.post(
            f"rf/task/{self.task_id}/submit",
            {
                "solverVersion": solver_version,
                "protocolVersion": protocol_version,
                "workerGroup": worker_group,
            },
        )

    def abort(self) -> requests.Response:
        """Abort the current task on the server."""
        if not self.task_id:
            raise ValueError("Batch id not found.")
        return http.put(f"rf/task/{self.task_id}/abort", {})


class TaskFactory:
    """Factory for obtaining the correct task subclass."""

    _REGISTRY: dict[str, str] = {}

    @classmethod
    def reset(cls) -> None:
        """Clear the cached task kind registry (used in tests)."""
        cls._REGISTRY.clear()

    @classmethod
    def register(cls, task_id: str, kind: str) -> None:
        cls._REGISTRY[task_id] = kind

    @classmethod
    def get(cls, task_id: str, verbose: bool = True) -> WebTask:
        kind = cls._REGISTRY.get(task_id)
        if kind == "batch":
            return BatchTask.get(task_id, verbose=verbose)
        if kind == "simulation":
            task = SimulationTask.get(task_id, verbose=verbose)
            return task
        if WebTask.is_batch(task_id):
            cls.register(task_id, "batch")
            return BatchTask.get(task_id, verbose=verbose)
        task = SimulationTask.get(task_id, verbose=verbose)
        if task:
            cls.register(task_id, "simulation")
        return task



"""Defines nonlinear models and specifications"""

from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Optional, Union

import autograd.numpy as np
import pydantic.v1 as pd

from tidy3d.constants import MICROMETER, SECOND, VOLT, WATT
from tidy3d.exceptions import SetupError, ValidationError

from .base import Tidy3dBaseModel

if TYPE_CHECKING:
    from .medium import AbstractMedium

# cap on number of nonlinear iterations
NONLINEAR_MAX_NUM_ITERS = 100
NONLINEAR_DEFAULT_NUM_ITERS = 5


class NonlinearModel(ABC, Tidy3dBaseModel):
    """Abstract model for a nonlinear material response.
    Used as part of a :class:`.NonlinearSpec`."""

    def _validate_medium_type(self, medium: AbstractMedium) -> None:
        """Check that the model is compatible with the medium."""
        from .medium import (
            AbstractCustomMedium,
            CustomDispersiveMedium,
            CustomMedium,
            DispersiveMedium,
            Medium,
        )

        if medium.is_time_modulated:
            raise ValidationError(
                f"'NonlinearModel' of class '{type(self).__name__}' is not currently supported "
                f"for time-modulated medium class '{type(medium).__name__}'."
            )
        if isinstance(medium, AbstractCustomMedium) and not medium.is_isotropic:
            raise ValidationError(
                f"'NonlinearModel' of class '{type(self).__name__}' is not currently supported "
                f"for anisotropic medium class '{type(medium).__name__}'."
            )
        if not isinstance(medium, (Medium, DispersiveMedium, CustomMedium, CustomDispersiveMedium)):
            raise ValidationError(
                f"'NonlinearModel' of class '{type(self).__name__}' is not currently supported "
                f"for medium class '{type(medium).__name__}'."
            )

    def _validate_medium(self, medium: AbstractMedium) -> None:
        """Any additional validation that depends on the medium"""

    def _validate_medium_freqs(self, medium: AbstractMedium, freqs: list[pd.PositiveFloat]) -> None:
        """Any additional validation that depends on the central frequencies of the sources."""

    @property
    def complex_fields(self) -> bool:
        """Whether the model uses complex fields."""
        return False

    @property
    def aux_fields(self) -> list[str]:
        """List of available aux fields in this model."""
        return []


class NonlinearSusceptibility(NonlinearModel):
    """Model for an instantaneous nonlinear chi3 susceptibility.
    The expression for the instantaneous nonlinear polarization is given below.

    Notes
    -----

        This model uses real time-domain fields, so :math:`\\chi_3` must be real.

        .. math::

            P_{NL} = \\varepsilon_0 \\chi_3 |E|^2 E

        The nonlinear constitutive relation is solved iteratively; it may not converge
        for strong nonlinearities. Increasing :attr:`tidy3d.NonlinearSpec.num_iters` can
        help with convergence.

        For complex fields (e.g. when using Bloch boundary conditions), the nonlinearity
        is applied separately to the real and imaginary parts, so that the above equation
        holds when both :math:`E` and :math:`P_{NL}` are replaced by their real or imaginary parts.
        The nonlinearity is only applied to the real-valued fields since they are the
        physical fields.

        Different field components do not interact nonlinearly. For example,
        when calculating :math:`P_{NL, x}`, we approximate :math:`|E|^2 \\approx |E_x|^2`.
        This approximation is valid when the :math:`E` field is predominantly polarized along one
        of the ``x``, ``y``, or ``z`` axes.

        .. TODO add links to notebooks here.

    Example
    -------
    >>> nonlinear_susceptibility = NonlinearSusceptibility(chi3=1)
    """

    chi3: float = pd.Field(
        0,
        title="Chi3",
        description=":math:`\\chi_3` nonlinear susceptibility.",
        units=f"{MICROMETER}^2 / {VOLT}^2",
    )

    numiters: pd.PositiveInt = pd.Field(
        None,
        title="Number of iterations",
        description="Deprecated. The old usage ``nonlinear_spec=model`` with ``model.numiters`` "
        "is deprecated and will be removed in a future release. The new usage is "
        "``nonlinear_spec=NonlinearSpec(models=[model], num_iters=num_iters)``. Under the new "
        "usage, this parameter is ignored, and ``NonlinearSpec.num_iters`` is used instead.",
    )

    @pd.validator("numiters", always=True)
    def _validate_numiters(cls, val):
        """Check that numiters is not too large."""
        if val is None:
            return val
        if val > NONLINEAR_MAX_NUM_ITERS:
            raise ValidationError(
                "'NonlinearSusceptibility.numiters' must be less than "
                f"{NONLINEAR_MAX_NUM_ITERS}, currently {val}."
            )
        return val


class TwoPhotonAbsorption(NonlinearModel):
    """Model for two-photon absorption (TPA) nonlinearity which gives an intensity-dependent
    absorption of the form :math:`\\alpha = \\alpha_0 + \\beta I`.
    Also includes free-carrier absorption (FCA) and free-carrier plasma dispersion (FCPD) effects.
    The expression for the nonlinear polarization is given below.

    Notes
    -----

        This model uses real time-domain fields, so :math:`\\beta` must be real.

        .. math::

            P_{NL} = P_{TPA} + P_{FCA} + P_{FCPD} \\\\
            P_{TPA} = -\\frac{4}{3}\\frac{c_0^2 \\varepsilon_0^2 n_0^2 \\beta}{2 i \\omega} |E|^2 E \\\\
            P_{FCA} = -\\frac{c_0 \\varepsilon_0 n_0 \\sigma N_f}{i \\omega} E \\\\
            \\frac{dN_f}{dt} = \\frac{8}{3}\\frac{c_0^2 \\varepsilon_0^2 n_0^2 \\beta}{8 q_e \\hbar \\omega} |E|^4 - \\frac{N_f}{\\tau} \\\\
            N_e = N_h = N_f \\\\
            P_{FCPD} = \\varepsilon_0 2 n_0 \\Delta n (N_f) E \\\\
            \\Delta n (N_f) = (c_e N_e^{e_e} + c_h N_h^{e_h})

        In these equations, :math:`n_0` means the real part of the linear
        refractive index of the medium.

        The nonlinear constitutive relation is solved iteratively; it may not converge
        for strong nonlinearities. Increasing :attr:`tidy3d.NonlinearSpec.num_iters` can
        help with convergence.

        For complex fields (e.g. when using Bloch boundary conditions), the nonlinearity
        is applied separately to the real and imaginary parts, so that the above equation
        holds when both :math:`E` and :math:`P_{NL}` are replaced by their real or imaginary parts.
        The nonlinearity is only applied to the real-valued fields since they are the
        physical fields.

        Different field components do not interact nonlinearly. For example,
        when calculating :math:`P_{NL, x}`, we approximate :math:`|E|^2 \\approx |E_x|^2`.
        This approximation is valid when the :math:`E` field is predominantly polarized along one
        of the ``x``, ``y``, or ``z`` axes.

        The implementation is described in::

            N. Suzuki, "FDTD Analysis of Two-Photon Absorption and Free-Carrier Absorption in Si
            High-Index-Contrast Waveguides," J. Light. Technol. 25, 9 (2007).

        .. TODO add links to notebooks here.

    Example
    -------
    >>> tpa_model = TwoPhotonAbsorption(beta=1)
    """

    beta: float = pd.Field(
        0,
        title="TPA coefficient",
        description="Coefficient for two-photon absorption (TPA).",
        units=f"{MICROMETER} / {WATT}",
    )

    tau: pd.NonNegativeFloat = pd.Field(
        0,
        title="Carrier lifetime",
        description="Lifetime for the free carriers created by two-photon absorption (TPA).",
        units=f"{SECOND}",
    )

    sigma: pd.NonNegativeFloat = pd.Field(
        0,
        title="FCA cross section",
        description="Total cross section for free-carrier absorption (FCA). "
        "Contains contributions from electrons and from holes.",
        units=f"{MICROMETER}^2",
    )
    e_e: pd.NonNegativeFloat = pd.Field(
        1,
        title="Electron exponent",
        description="Exponent for the free electron refractive index shift in the free-carrier plasma dispersion (FCPD).",
    )
    e_h: pd.NonNegativeFloat = pd.Field(
        1,
        title="Hole exponent",
        description="Exponent for the free hole refractive index shift in the free-carrier plasma dispersion (FCPD).",
    )
    c_e: float = pd.Field(
        0,
        title="Electron coefficient",
        description="Coefficient for the free electron refractive index shift in the free-carrier plasma dispersion (FCPD).",
        units=f"{MICROMETER}^(3 e_e)",
    )
    c_h: float = pd.Field(
        0,
        title="Hole coefficient",
        description="Coefficient for the free hole refractive index shift in the free-carrier plasma dispersion (FCPD).",
        units=f"{MICROMETER}^(3 e_h)",
    )

    n0: Optional[float] = pd.Field(
        None,
        title="Linear refractive index",
        description="Real linear refractive index of the medium, computed for instance using "
        "'medium.nk_model'. If not provided, it is calculated automatically using the central "
        "frequencies of the simulation sources (as long as these are all equal).",
    )

    freq0: Optional[pd.PositiveFloat] = pd.Field(
        None,
        title="Central frequency",
        description="Central frequency, used to calculate the energy of the free-carriers "
        "excited by two-photon absorption. If not provided, it is obtained automatically "
        "from the simulation sources (as long as these are all equal).",
    )

    def _validate_medium(self, medium: AbstractMedium) -> None:
        """Check that the model is compatible with the medium."""
        beta = self.beta
        if not medium.allow_gain and beta < 0:
            raise ValidationError(
                "A passive medium must have 'beta >= 0' in "
                f"'TwoPhotonAbsorption', given 'beta={beta}'. "
                "To simulate a gain medium, please set "
                "'allow_gain=True' in the medium class. Caution: "
                "simulations containing gain medium are unstable, "
                "and are likely to diverge."
            )

    @property
    def aux_fields(self) -> list[str]:
        """List of available aux fields in this model."""
        if self.tau == 0:
            return []
        return ["Nfx", "Nfy", "Nfz"]


class KerrNonlinearity(NonlinearModel):
    """Model for Kerr nonlinearity which gives an intensity-dependent refractive index
    of the form :math:`n = n_0 + n_2 I`. The expression for the nonlinear polarization
    is given below.

    Notes
    -----

        This model uses real time-domain fields, so :math:`\\n_2` must be real.

        This model is equivalent to a :class:`.NonlinearSusceptibility`; the
        relation between the parameters is given below.

        .. math::

            P_{NL} = \\varepsilon_0 \\chi_3 |E|^2 E \\\\
            n_2 = \\frac{3}{4 n_0^2 \\varepsilon_0 c_0} \\chi_3

        In these equations, :math:`n_0` means the real part of the linear
        refractive index of the medium.

        To simulate nonlinear loss, consider instead using a :class:`.TwoPhotonAbsorption`
        model, which implements a more physical dispersive loss of the form
        :math:`\\chi_{TPA} = i \\frac{c_0 n_0 \\beta}{\\omega} I`.

        The nonlinear constitutive relation is solved iteratively; it may not converge
        for strong nonlinearities. Increasing :attr:`tidy3d.NonlinearSpec.num_iters` can
        help with convergence.

        For complex fields (e.g. when using Bloch boundary conditions), the nonlinearity
        is applied separately to the real and imaginary parts, so that the above equation
        holds when both :math:`E` and :math:`P_{NL}` are replaced by their real or imaginary parts.
        The nonlinearity is only applied to the real-valued fields since they are the
        physical fields.

        Different field components do not interact nonlinearly. For example,
        when calculating :math:`P_{NL, x}`, we approximate :math:`|E|^2 \\approx |E_x|^2`.
        This approximation is valid when the :math:`E` field is predominantly polarized along one
        of the ``x``, ``y``, or ``z`` axes.

        .. TODO add links to notebooks here.

    Example
    -------
    >>> kerr_model = KerrNonlinearity(n2=1)
    """

    n2: float = pd.Field(
        0,
        title="Nonlinear refractive index",
        description="Nonlinear refractive index in the Kerr nonlinearity.",
        units=f"{MICROMETER}^2 / {WATT}",
    )

    n0: Optional[float] = pd.Field(
        None,
        title="Complex linear refractive index",
        description="Complex linear refractive index of the medium, computed for instance using "
        "'medium.nk_model'. If not provided, it is calculated automatically using the central "
        "frequencies of the simulation sources (as long as these are all equal).",
    )


NonlinearModelType = Union[NonlinearSusceptibility, TwoPhotonAbsorption, KerrNonlinearity]


class NonlinearSpec(ABC, Tidy3dBaseModel):
    """Abstract specification for adding nonlinearities to a medium.

    Note
    ----
    The nonlinear constitutive relation is solved iteratively; it may not converge
    for strong nonlinearities. Increasing ``num_iters`` can help with convergence.

    Example
    -------
    >>> from tidy3d import Medium
    >>> nonlinear_susceptibility = NonlinearSusceptibility(chi3=1)
    >>> nonlinear_spec = NonlinearSpec(models=[nonlinear_susceptibility])
    >>> medium = Medium(permittivity=2, nonlinear_spec=nonlinear_spec)
    """

    models: tuple[NonlinearModelType, ...] = pd.Field(
        (),
        title="Nonlinear models",
        description="The nonlinear models present in this nonlinear spec. "
        "Nonlinear models of different types are additive. "
        "Multiple nonlinear models of the same type are not allowed.",
    )

    num_iters: pd.PositiveInt = pd.Field(
        NONLINEAR_DEFAULT_NUM_ITERS,
        title="Number of iterations",
        description="Number of iterations for solving nonlinear constitutive relation.",
    )

    @pd.validator("models", always=True)
    def _no_duplicate_models(cls, val):
        """Ensure each type of model appears at most once."""
        if val is None:
            return val
        models = [model.__class__ for model in val]
        models_unique = set(models)
        if len(models) != len(models_unique):
            raise ValidationError(
                "Multiple 'NonlinearModels' of the same type "
                "were found in a single 'NonlinearSpec'. Please ensure that "
                "each type of 'NonlinearModel' appears at most once in a single 'NonlinearSpec'."
            )
        return val

    @pd.validator("num_iters", always=True)
    def _validate_num_iters(cls, val, values):
        """Check that num_iters is not too large."""
        if val > NONLINEAR_MAX_NUM_ITERS:
            raise ValidationError(
                "'NonlinearSpec.num_iters' must be less than "
                f"{NONLINEAR_MAX_NUM_ITERS}, currently {val}."
            )
        return val

    @property
    def aux_fields(self) -> list[str]:
        """List of available aux fields in all present models."""
        fields = []
        for model in self.models:
            fields += model.aux_fields
        return fields

    @pd.validator("models", always=True)
    def _consistent_models(cls, val):
        """Ensure that parameters shared between models are consistent."""
        if val is None:
            return val
        n0 = None
        for model in val:
            if isinstance(model, (KerrNonlinearity, TwoPhotonAbsorption)):
                if model.n0 is not None:
                    if n0 is not None and not np.isclose(model.n0, n0):
                        raise SetupError(
                            f"Nonlinear models must have consistent 'n0'. Given {model.n0} and {n0}."
                        )
                    n0 = model.n0
        return val



"""Extrapolation into low frequencies specification."""

from __future__ import annotations

from typing import Optional

import pydantic.v1 as pydantic

from tidy3d.components.base import Tidy3dBaseModel


class AbstractLowFrequencySmoothingSpec(Tidy3dBaseModel):
    """Abstract base class for low frequency smoothing specifications."""

    min_sampling_time: pydantic.NonNegativeFloat = pydantic.Field(
        1.0,
        title="Minimum Sampling Time (periods)",
        description="The minimum simulation time in periods of the corresponding frequency for which frequency domain results will be used to fit the polynomial for the low frequency extrapolation. "
        "Results below this threshold will be completely discarded.",
    )

    max_sampling_time: pydantic.NonNegativeFloat = pydantic.Field(
        5.0,
        title="Maximum Sampling Time (periods)",
        description="The maximum simulation time in periods of the corresponding frequency for which frequency domain results will be used to fit the polynomial for the low frequency extrapolation. "
        "Results above this threshold will be not be modified.",
    )

    order: int = pydantic.Field(
        1,
        title="Extrapolation Order",
        description="The order of the polynomial to use for the low frequency extrapolation.",
        ge=0,
        le=3,
    )

    max_deviation: Optional[float] = pydantic.Field(
        0.5,
        title="Maximum Deviation",
        description="The maximum deviation (in fraction of the trusted values) to allow for the low frequency smoothing.",
        ge=0,
    )

    @pydantic.root_validator(skip_on_failure=True)
    def _validate_sampling_times(cls, values):
        min_sampling_time = values.get("min_sampling_time")
        max_sampling_time = values.get("max_sampling_time")
        if min_sampling_time >= max_sampling_time:
            raise ValueError(
                "The minimum sampling time must be less than the maximum sampling time."
            )
        return values


class LowFrequencySmoothingSpec(AbstractLowFrequencySmoothingSpec):
    """Specifies the low frequency smoothing parameters for the simulation.
    This specification affects only results recorded in mode monitors. Specifically, the mode decomposition data
    for frequencies for which the total simulation time in units of the corresponding period (T = 1/f) is less than
    the specified minimum sampling time will be overridden by extrapolation from the data in the trusted frequency range.
    The trusted frequency range is defined in terms of minimum and maximum sampling times (the total simulation time divided by the corresponding period).
    Example
    -------
    >>> low_freq_smoothing = LowFrequencySmoothingSpec(
    ...     min_sampling_time=3,
    ...     max_sampling_time=6,
    ...     order=1,
    ...     max_deviation=0.5,
    ...     monitors=("monitor1", "monitor2"),
    ... )
    """

    monitors: tuple[str, ...] = pydantic.Field(
        ...,
        title="Monitors",
        description="The names of monitors to which low frequency smoothing will be applied.",
    )

    @pydantic.validator("monitors", always=True)
    def _validate_monitors(cls, val, values):
        """Validate the monitors list is not empty."""
        if not val:
            raise ValueError("The monitors list must not be empty.")
        return val



"""Backwards compatibility - import from tidy3d.components.microwave.path_integrals.integrals instead."""

from __future__ import annotations

from tidy3d.components.data.data_array import (
    CurrentIntegralResultType,
    IntegralResultType,
    VoltageIntegralResultType,
    _make_base_result_data_array,
    _make_current_data_array,
    _make_voltage_data_array,
)
from tidy3d.components.microwave.path_integrals.integrals.base import (
    Custom2DPathIntegral,
)
from tidy3d.components.microwave.path_integrals.integrals.current import (
    CompositeCurrentIntegral,
    Custom2DCurrentIntegral,
)
from tidy3d.components.microwave.path_integrals.integrals.voltage import (
    Custom2DVoltageIntegral,
)

__all__ = [
    "CompositeCurrentIntegral",
    "CurrentIntegralResultType",
    "Custom2DCurrentIntegral",
    "Custom2DPathIntegral",
    "Custom2DVoltageIntegral",
    "IntegralResultType",
    "VoltageIntegralResultType",
    "_make_base_result_data_array",
    "_make_current_data_array",
    "_make_voltage_data_array",
]



"""Provides lowest level, user-facing interface to server."""

from __future__ import annotations

import json
import tempfile
import time
from os import PathLike
from pathlib import Path
from typing import Callable, Literal, Optional, Union

from requests import HTTPError
from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn, TimeElapsedColumn

from tidy3d.components.medium import AbstractCustomMedium
from tidy3d.components.mode.mode_solver import ModeSolver
from tidy3d.components.mode.simulation import ModeSimulation
from tidy3d.components.types.workflow import WorkflowDataType, WorkflowType
from tidy3d.config import config
from tidy3d.exceptions import WebError
from tidy3d.log import get_logging_console, log
from tidy3d.web.api.states import (
    ALL_POST_VALIDATE_STATES,
    END_STATES,
    ERROR_STATES,
    MAX_STEPS,
    STATE_PROGRESS_PERCENTAGE,
    status_to_stage,
)
from tidy3d.web.cache import CacheEntry, _store_mode_solver_in_cache, resolve_local_cache
from tidy3d.web.core.account import Account
from tidy3d.web.core.constants import (
    CM_DATA_HDF5_GZ,
    MODE_DATA_HDF5_GZ,
    MODE_FILE_HDF5_GZ,
    MODELER_FILE_HDF5_GZ,
    SIM_FILE_HDF5,
    SIM_FILE_HDF5_GZ,
    SIMULATION_DATA_HDF5_GZ,
    TaskId,
)
from tidy3d.web.core.task_core import (
    BatchDetail,
    BatchTask,
    Folder,
    SimulationTask,
    TaskFactory,
    WebTask,
)
from tidy3d.web.core.task_info import ChargeType, TaskInfo
from tidy3d.web.core.types import PayType, TaskType

from .connect_util import REFRESH_TIME, get_grid_points_str, get_time_steps_str, wait_for_connection
from .tidy3d_stub import Tidy3dStub, Tidy3dStubData

# time between checking run status
RUN_REFRESH_TIME = 1.0

# file names when uploading to S3
SIM_FILE_JSON = "simulation.json"

# not all solvers are supported yet in GUI
GUI_SUPPORTED_TASK_TYPES = ["FDTD", "MODE_SOLVER", "HEAT", "TERMINAL_CM"]

# if a solver is in beta stage, cost is subject to change
BETA_TASK_TYPES = ["HEAT", "EME", "HEAT_CHARGE", "VOLUME_MESH"]

# map task_type to solver name for display
SOLVER_NAME = {
    "FDTD": "FDTD",
    "MODE_SOLVER": "Mode",
    "MODE": "Mode",
    "EME": "EME",
    "HEAT": "Heat",
    "HEAT_CHARGE": "HeatCharge",
    "VOLUME_MESH": "VolumeMesher",
}


def _get_url(task_id: str) -> str:
    """Get the URL for a task on our server."""
    return _build_website_url(f"workbench?taskId={task_id}")


def _get_folder_url(folder_id: str) -> str:
    """Get the URL for a task folder on our server."""
    return _build_website_url(f"folders/{folder_id}")


def _get_url_rf(resource_id: str) -> str:
    """Get the RF GUI URL for a modeler/batch group."""
    return _build_website_url(f"rf?taskId={resource_id}")


def _build_website_url(path: str) -> str:
    base = str(config.web.website_endpoint or "")
    if not path:
        return base
    return "/".join([base.rstrip("/"), str(path).lstrip("/")])


def _batch_detail_error(resource_id: str) -> Optional[WebError]:
    """Processes a failed batch job to generate a detailed error.

    This function inspects the status of a batch detail object. If the status
    indicates an error, it logs the failure and constructs a specific `WebError`
    object to be returned. For validation failures, it parses and aggregates
    detailed error messages from each subtask.

    Args:
        resource_id (str): The identifier of the batch resource that failed.

    Raises:
        An instance of ``WebError`` if the batch failed.
    """

    # TODO: test properly
    try:
        batch = BatchTask.get(resource_id)
        batch_detail = batch.detail()
        status = batch_detail.status.lower()
    except Exception as e:
        log.error(f"Could not retrieve batch details for '{resource_id}': {e}")
        raise WebError(f"Failed to retrieve status for batch '{resource_id}'.") from e

    if status not in ERROR_STATES:
        return

    if hasattr(batch_detail, "validateErrors") and batch_detail.validateErrors:
        try:
            error_details = []
            for key, error_str in batch_detail.validateErrors.items():
                msg = f"- Subtask '{key}' failed: {error_str}"
                error_details.append(msg)

            details_string = "\n".join(error_details)
            full_error_msg = (
                "One or more subtasks failed validation. Please fix the component modeler "
                "configuration.\n"
                f"Details:\n{details_string}"
            )
        except Exception as e:
            raise WebError(
                "One or more subtasks failed validation. Failed to parse validation errors."
            ) from e
        raise WebError(full_error_msg)

    # Handle all other generic error states
    else:
        error_msg = (
            f"Batch '{resource_id}' failed with status '{status}'. Check server "
            "logs for details or contact customer support."
        )
        raise WebError(error_msg)


def _copy_simulation_data_from_cache_entry(entry: CacheEntry, path: PathLike) -> bool:
    """
    Copy cached simulation data from a cache entry to a specified path.

    Parameters
    ----------
    entry : CacheEntry
        The cache entry containing simulation data and metadata.
    path : PathLike
        The target directory or file path where the cached data should be materialized.

    Returns
    -------
    bool
        True if the cached simulation data was successfully copied, False otherwise.
    """
    if entry is not None:
        try:
            entry.materialize(Path(path))
            return True
        except Exception:
            return False
    return False


def restore_simulation_if_cached(
    simulation: WorkflowType,
    path: Optional[PathLike] = None,
    reduce_simulation: Literal["auto", True, False] = "auto",
    verbose: bool = True,
) -> tuple[Optional[PathLike], Optional[TaskId]]:
    """
    Attempt to restore simulation data from a local cache entry, if available.

    Parameters
    ----------
    simulation : WorkflowType
        The simulation or workflow object for which cached data may exist.
    path : Optional[PathLike] = None
        Optional path where the cached data should be copied. If not provided,
        the path from the cache entry will be used.
    reduce_simulation : Literal["auto", True, False] = "auto"
        Whether to reduce the simulation for cache lookup. If "auto", reduction is applied
        only when applicable (e.g., for mode solvers).
    verbose : bool = True
        If True, logs a message including a link to the cached task in the web UI.

    Returns
    -------
    Optional[PathLike]
        The path to the restored simulation data if found in cache, otherwise None. If no path is specified, the cache entry path is returned, otherwise the given path is returned.
    Optional[TaskId]
        The original task id of the restored simulation data.
    """
    simulation_cache = resolve_local_cache()
    retrieved_simulation_path = None
    cached_task_id = None
    if simulation_cache is not None:
        sim_for_cache = simulation
        if isinstance(simulation, (ModeSolver, ModeSimulation)):
            sim_for_cache = get_reduced_simulation(simulation, reduce_simulation)
        entry = simulation_cache.try_fetch(simulation=sim_for_cache, verbose=verbose)
        if entry is not None:
            if path is not None:
                copied = _copy_simulation_data_from_cache_entry(entry, path)
                if copied:
                    retrieved_simulation_path = path
            else:
                retrieved_simulation_path = entry.artifact_path
            cached_task_id = entry.metadata.get("task_id")
            cached_workflow_type = entry.metadata.get("workflow_type")
            if cached_task_id is not None and cached_workflow_type is not None and verbose:
                console = get_logging_console()
                url, _ = _get_task_urls(cached_workflow_type, cached_task_id)
                console.log(
                    f"Loading simulation from local cache. View cached task using web UI at [link={url}]'{url}'[/link]."
                )
    return retrieved_simulation_path, cached_task_id


def load_simulation_if_cached(
    simulation: WorkflowType,
    path: Optional[PathLike] = None,
    reduce_simulation: Literal["auto", True, False] = "auto",
    verbose: bool = True,
) -> Optional[WorkflowDataType]:
    """
    Load simulation results directly from the local cache, if available.

    Parameters
    ----------
    simulation : WorkflowType
        The simulation or workflow object to check for cached results.
    path : Optional[PathLike] = None
        Optional path to which cached data should be restored before loading.
    reduce_simulation : Literal["auto", True, False] = "auto"
        Whether to use a reduced simulation when checking the cache. If "auto",
        reduction is applied automatically for mode solvers.
    verbose : bool = True
        If True, logs a message including a link to the cached task in the web UI on loading.

    Returns
    -------
    Optional[WorkflowDataType]
        The loaded simulation data if found in cache, otherwise None.
    """
    restored_path, _ = restore_simulation_if_cached(
        simulation, path, reduce_simulation, verbose=verbose
    )
    if restored_path is not None:
        data = load(
            task_id=None,
            path=str(restored_path),
            verbose=verbose,
        )
        if isinstance(simulation, ModeSolver):
            simulation._patch_data(data=data)
        return data
    else:
        return None


@wait_for_connection
def run(
    simulation: WorkflowType,
    task_name: Optional[str] = None,
    folder_name: str = "default",
    path: PathLike = "simulation_data.hdf5",
    callback_url: Optional[str] = None,
    verbose: bool = True,
    progress_callback_upload: Optional[Callable[[float], None]] = None,
    progress_callback_download: Optional[Callable[[float], None]] = None,
    solver_version: Optional[str] = None,
    worker_group: Optional[str] = None,
    simulation_type: str = "tidy3d",
    parent_tasks: Optional[list[str]] = None,
    reduce_simulation: Literal["auto", True, False] = "auto",
    pay_type: Union[PayType, str] = PayType.AUTO,
    priority: Optional[int] = None,
    lazy: bool = False,
) -> WorkflowDataType:
    """
    Submits a :class:`.Simulation` to server, starts running, monitors progress, downloads,
    and loads results as a :class:`.WorkflowDataType` object.

    Parameters
    ----------
    simulation : Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`]
        Simulation to upload to server.
    task_name : Optional[str] = None
        Name of task. If not provided, a default name will be generated.
    folder_name : str = "default"
        Name of folder to store task on web UI.
    path : PathLike = "simulation_data.hdf5"
        Path to download results file (.hdf5), including filename.
    callback_url : str = None
        Http PUT url to receive simulation finish event. The body content is a json file with
        fields ``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    simulation_type : str = "tidy3d"
        Type of simulation being uploaded.
    progress_callback_upload : Callable[[float], None] = None
        Optional callback function called when uploading file with ``bytes_in_chunk`` as argument.
    progress_callback_download : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.
    solver_version: str = None
        target solver version.
    worker_group: str = None
        worker group
    reduce_simulation : Literal["auto", True, False] = "auto"
        Whether to reduce structures in the simulation to the simulation domain only. Note: currently only implemented for the mode solver.
    pay_type: Union[PayType, str] = PayType.AUTO
        Which method to pay the simulation.
    priority: int = None
        Priority of the simulation in the Virtual GPU (vGPU) queue (1 = lowest, 10 = highest).
        It affects only simulations from vGPU licenses and does not impact simulations using FlexCredits.
    lazy : bool = False
        Whether to load the actual data (``lazy=False``) or return a proxy that loads
        the data when accessed (``lazy=True``).

    Returns
    -------
    Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`]
        Object containing solver results for the supplied simulation.

    Notes
    -----

        Submitting a simulation to our cloud server is very easily done by a simple web API call.

        .. code-block:: python

            sim_data = tidy3d.web.api.webapi.run(simulation, task_name='my_task', path='out/data.hdf5')

        The :meth:`tidy3d.web.api.webapi.run()` method shows the simulation progress by default.  When uploading a
        simulation to the server without running it, you can use the :meth:`tidy3d.web.api.webapi.monitor`,
        :meth:`tidy3d.web.api.container.Job.monitor`, or :meth:`tidy3d.web.api.container.Batch.monitor` methods to
        display the progress of your simulation(s).

    Examples
    --------

        To access the original :class:`.Simulation` object that created the simulation data you can use:

        .. code-block:: python

            # Run the simulation.
            sim_data = web.run(simulation, task_name='task_name', path='out/sim.hdf5')

            # Get a copy of the original simulation object.
            sim_copy = sim_data.simulation

    See Also
    --------

    :meth:`tidy3d.web.api.webapi.monitor`
        Print the real time task progress until completion.

    :meth:`tidy3d.web.api.container.Job.monitor`
        Monitor progress of running :class:`Job`.

    :meth:`tidy3d.web.api.container.Batch.monitor`
        Monitor progress of each of the running tasks.
    """
    restored_path, _ = restore_simulation_if_cached(
        simulation=simulation,
        path=path,
        reduce_simulation=reduce_simulation,
        verbose=verbose,
    )

    if not restored_path:
        task_id = upload(
            simulation=simulation,
            task_name=task_name,
            folder_name=folder_name,
            callback_url=callback_url,
            verbose=verbose,
            progress_callback=progress_callback_upload,
            simulation_type=simulation_type,
            parent_tasks=parent_tasks,
            solver_version=solver_version,
            reduce_simulation=reduce_simulation,
        )
        start(
            task_id,
            solver_version=solver_version,
            worker_group=worker_group,
            pay_type=pay_type,
            priority=priority,
        )
        monitor(task_id, verbose=verbose)
    else:
        task_id = None

    data = load(
        task_id=task_id,
        path=path,
        verbose=verbose,
        progress_callback=progress_callback_download,
        lazy=lazy,
    )

    if isinstance(simulation, ModeSolver):
        if task_id is not None:
            _store_mode_solver_in_cache(task_id, simulation, data, path)
        simulation._patch_data(data=data)

    return data


def _get_task_urls(
    task_type: str,
    resource_id: str,
    folder_id: Optional[str] = None,
    group_id: Optional[str] = None,
) -> tuple[str, Optional[str]]:
    """Log task and folder links to the web UI."""
    if task_type in ["RF", "TERMINAL_CM", "MODAL_CM"]:
        url = _get_url_rf(group_id or resource_id)
    else:
        url = _get_url(resource_id)

    if folder_id is not None:
        folder_url = _get_folder_url(folder_id)
    else:
        folder_url = None
    return url, folder_url


@wait_for_connection
def upload(
    simulation: WorkflowType,
    task_name: Optional[str] = None,
    folder_name: str = "default",
    callback_url: Optional[str] = None,
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
    simulation_type: str = "tidy3d",
    parent_tasks: Optional[list[str]] = None,
    source_required: bool = True,
    solver_version: Optional[str] = None,
    reduce_simulation: Literal["auto", True, False] = "auto",
    verbose_estimate_cost: Optional[bool] = None,
) -> TaskId:
    """
    Upload simulation to server, but do not start running :class:`.Simulation`.

    Parameters
    ----------
    simulation : Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`]
        Simulation to upload to server.
    task_name : Optional[str]
        Name of task. If not provided, a default name will be generated.
    folder_name : str
        Name of folder to store task on web UI
    callback_url : str = None
        Http PUT url to receive simulation finish event. The body content is a json file with
        fields ``{'id', 'status', 'name', 'workUnit', 'solverVersion'}``.
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when uploading file with ``bytes_in_chunk`` as argument.
    simulation_type : str = "tidy3d"
        Type of simulation being uploaded.
    parent_tasks : List[str]
        List of related task ids.
    source_required: bool = True
        If ``True``, simulations without sources will raise an error before being uploaded.
    solver_version: str = None
        target solver version.
    reduce_simulation: Literal["auto", True, False] = "auto"
        Whether to reduce structures in the simulation to the simulation domain only. Note: currently only implemented for the mode solver.
    verbose_estimate_cost : Optional[bool] = None
        Determines if cost estimation should be printed. If ``None``, defaults to ``verbose`` argument.

    Returns
    -------
    str
        Unique identifier of task on server.


    Notes
    -----

        Once you've created a ``job`` object using :class:`tidy3d.web.api.container.Job`, you can upload it to our servers with:

        .. code-block:: python

            web.upload(simulation, task_name="task_name", verbose=verbose)

        It will not run until you explicitly tell it to do so with :meth:`tidy3d.web.api.webapi.start`.

    """
    console = get_logging_console() if verbose else None

    if isinstance(simulation, (ModeSolver, ModeSimulation)):
        simulation = get_reduced_simulation(simulation, reduce_simulation)

    stub = Tidy3dStub(simulation=simulation)
    stub.validate_pre_upload(source_required=source_required)
    log.debug("Creating task.")

    if task_name is None:
        task_name = stub.get_default_task_name()

    task_type = stub.get_type()

    task = WebTask.create(
        task_type,
        task_name,
        folder_name,
        callback_url,
        simulation_type,
        parent_tasks,
        "Gz",
    )

    group_id = getattr(task, "groupId", None)
    resource_id = task.task_id

    if verbose:
        console.log(
            f"Created task '{task_name}' with resource_id '{resource_id}' and task_type '{task_type}'."
        )
        if task_type in BETA_TASK_TYPES:
            solver_name = SOLVER_NAME[task_type]
            console.log(
                f"Tidy3D's {solver_name} solver is currently in the beta stage. "
                f"Cost of {solver_name} simulations is subject to change in the future."
            )
        if task_type in GUI_SUPPORTED_TASK_TYPES:
            url, folder_url = _get_task_urls(task_type, resource_id, task.folder_id, group_id)
            console.log(f"View task using web UI at [link={url}]'{url}'[/link].")
            console.log(f"Task folder: [link={folder_url}]'{task.folder_name}'[/link].")

    remote_sim_file = SIM_FILE_HDF5_GZ
    if task_type == "MODE_SOLVER":
        remote_sim_file = MODE_FILE_HDF5_GZ
    elif task_type in ["RF", "TERMINAL_CM", "MODAL_CM"]:
        remote_sim_file = MODELER_FILE_HDF5_GZ

    task.upload_simulation(
        stub=stub,
        verbose=verbose,
        progress_callback=progress_callback,
        remote_sim_file=remote_sim_file,
    )

    verbose_estimate_cost = verbose if verbose_estimate_cost is None else verbose_estimate_cost
    estimate_cost(task_id=resource_id, solver_version=solver_version, verbose=verbose_estimate_cost)

    task.validate_post_upload(parent_tasks=parent_tasks)

    return resource_id


def get_reduced_simulation(
    simulation: WorkflowType, reduce_simulation: Literal["auto", True, False]
) -> WorkflowType:
    """
    Adjust the given simulation object based on the reduce_simulation parameter. Currently only
    implemented for the mode solver.

    Parameters
    ----------
    simulation : Simulation
        The simulation object to be potentially reduced.
    reduce_simulation : Literal["auto", True, False]
        Determines whether to reduce the simulation. If "auto", the function will decide based on
        the presence of custom mediums in the simulation.

    Returns
    -------
    Simulation
        The potentially reduced simulation object.
    """

    """
    TODO: This only works for the mode solver, which is also why `simulation.simulation.scene` is
    used below. After refactor to use the new ModeSimulation, it should be possible to put the call
    to this function outside of the MODE_SOLVER check in the upload function. We could implement
    dummy `reduced_simulation_copy` methods for the other solvers or also implement reductions
    there. Note that if we do the latter we may want to also modify the warning below to only
    happen if there are custom media *and* they extend beyond the simulation domain.
    """
    if reduce_simulation == "auto":
        if isinstance(simulation, ModeSimulation):
            sim_mediums = simulation.scene.mediums
        else:
            sim_mediums = simulation.simulation.scene.mediums
        contains_custom = any(isinstance(med, AbstractCustomMedium) for med in sim_mediums)
        reduce_simulation = contains_custom

        if reduce_simulation:
            log.warning(
                f"The {type(simulation)} object contains custom mediums. It will be "
                "automatically restricted to the solver domain to reduce data for uploading. "
                "To force uploading the original object use 'reduce_simulation=False'."
                " Setting 'reduce_simulation=True' will force simulation reduction in all cases and"
                " silence this warning."
            )
    if reduce_simulation:
        return simulation.reduced_simulation_copy
    return simulation


@wait_for_connection
def get_info(task_id: TaskId, verbose: bool = True) -> TaskInfo | BatchDetail:
    """Return information about a simulation task or a modeler batch.

    This function fetches details for a given task ID, automatically
    distinguishing between a standard simulation task and a modeler batch.

    Parameters
    ----------
    task_id : TaskId
        The unique identifier for the task or batch.
    verbose : bool = True
        If ``True`` (default), display progress bars and status updates.
        If ``False``, the function runs silently.

    Returns
    -------
    TaskInfo | BatchDetail
        A ``TaskInfo`` object for a standard simulation task, or a
        ``BatchDetail`` object for a modeler batch.

    Raises
    ------
    ValueError
        If no task is found for the given ``task_id``.
    """
    task = TaskFactory.get(task_id, verbose=verbose)
    if not task:
        raise ValueError("Task not found.")
    return task.detail()


@wait_for_connection
def start(
    task_id: TaskId,
    solver_version: Optional[str] = None,
    worker_group: Optional[str] = None,
    pay_type: Union[PayType, str] = PayType.AUTO,
    priority: Optional[int] = None,
) -> None:
    """Start running the simulation associated with task.

    Parameters
    ----------

    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    verbose : bool = True
        If ``True``, will print log messages, otherwise, will run silently.
    solver_version: str = None
        target solver version.
    worker_group: str = None
        worker group
    pay_type: Union[PayType, str] = PayType.AUTO
        Which method to pay the simulation
    priority: int = None
        Priority of the simulation in the Virtual GPU (vGPU) queue (1 = lowest, 10 = highest).
        It affects only simulations from vGPU licenses and does not impact simulations using FlexCredits.
    Note
    ----
    To monitor progress, can call :meth:`monitor` after starting simulation.
    """

    if priority is not None and (priority < 1 or priority > 10):
        raise ValueError("Priority must be between '1' and '10' if specified.")

    task = TaskFactory.get(task_id)
    if not task:
        raise ValueError("Task not found.")
    task.submit(
        solver_version=solver_version,
        worker_group=worker_group,
        pay_type=pay_type,
        priority=priority,
    )


@wait_for_connection
def get_run_info(task_id: TaskId) -> tuple[Optional[float], Optional[float]]:
    """Gets the % done and field_decay for a running task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.

    Returns
    -------
    perc_done : float
        Percentage of run done (in terms of max number of time steps).
        Is ``None`` if run info not available.
    field_decay : float
        Average field intensity normalized to max value (1.0).
        Is ``None`` if run info not available.
    """
    task = TaskFactory.get(task_id)
    if isinstance(task, BatchTask):
        raise NotImplementedError("Operation not implemented for modeler batches.")
    return task.get_running_info()


def _get_batch_detail_handle_error_status(batch: BatchTask) -> BatchDetail:
    """Get batch detail and raise error if status is in ERROR_STATES."""
    detail = batch.detail()
    status = detail.status.lower()
    if status in ERROR_STATES:
        _batch_detail_error(batch.task_id)
    return detail


def get_status(task_id: TaskId) -> str:
    """Get the status of a task. Raises an error if status is "error".

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    """
    task = TaskFactory.get(task_id)
    if isinstance(task, BatchTask):
        return _get_batch_detail_handle_error_status(task).status
    else:
        task_info = get_info(task_id)
        status = task_info.status
        if status == "visualize":
            return "success"
        if status in ERROR_STATES:
            try:
                # Try to obtain the error message
                task = SimulationTask(taskId=task_id)
                with tempfile.NamedTemporaryFile(suffix=".json") as tmp_file:
                    task.get_error_json(to_file=tmp_file.name)
                    with open(tmp_file.name) as f:
                        error_content = json.load(f)
                        error_msg = error_content["msg"]
            except Exception:
                # If the error message could not be obtained, raise a generic error message
                error_msg = "Error message could not be obtained, please contact customer support."

            raise WebError(f"Error running task {task_id}! {error_msg}")
    return status


def monitor(task_id: TaskId, verbose: bool = True, worker_group: Optional[str] = None) -> None:
    """
    Print the real time task progress until completion.

    Notes
    -----

        To monitor the simulation's progress and wait for its completion, use:

        .. code-block:: python

            tidy3d.web.api.webapi.monitor(job.task_id, verbose=verbose).

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.

    Note
    ----
    To load results when finished, may call :meth:`load`.
    """

    # Batch/modeler monitoring path
    task = TaskFactory.get(task_id)
    if isinstance(task, BatchTask):
        return _monitor_modeler_batch(task_id, verbose=verbose)

    console = get_logging_console() if verbose else None

    task_info = get_info(task_id)

    task_name = task_info.taskName

    task_type = task_info.taskType

    def get_estimated_cost() -> float:
        """Get estimated cost, if None, is not ready."""
        task_info = get_info(task_id)
        block_info = task_info.taskBlockInfo
        if block_info and block_info.chargeType == ChargeType.FREE:
            est_flex_unit = 0
            grid_points = block_info.maxGridPoints
            time_steps = block_info.maxTimeSteps
            grid_points_str = get_grid_points_str(grid_points)
            time_steps_str = get_time_steps_str(time_steps)
            console.log(
                f"You are running this simulation for FREE. Your current plan allows"
                f" up to {block_info.maxFreeCount} free non-concurrent simulations per"
                f" day (under {grid_points_str} grid points and {time_steps_str}"
                f" time steps)"
            )
        else:
            est_flex_unit = task_info.estFlexUnit
        return est_flex_unit

    def monitor_preprocess() -> None:
        """Periodically check the status."""
        status = get_status(task_id)
        while status not in END_STATES and status != "running":
            new_status = get_status(task_id)
            if new_status != status:
                status = new_status
                if verbose and status != "running":
                    console.log(f"status = {status}")
            time.sleep(REFRESH_TIME)

    status = get_status(task_id)

    if verbose:
        console.log(f"status = {status}")

    # already done
    if status in END_STATES:
        return

    # preprocessing
    if verbose:
        console.log(
            "To cancel the simulation, use 'web.abort(task_id)' or 'web.delete(task_id)' "
            "or abort/delete the task in the web "
            "UI. Terminating the Python script will not stop the job running on the cloud."
        )
        with console.status(f"[bold green]Waiting for '{task_name}'...", spinner="runner"):
            monitor_preprocess()
    else:
        monitor_preprocess()

    # if the estimated cost is ready, print it
    if verbose:
        get_estimated_cost()
        console.log("starting up solver")

    # while running but before the percentage done is available, keep waiting
    while get_run_info(task_id)[0] is None and get_status(task_id) == "running":
        time.sleep(REFRESH_TIME)

    # while running but percentage done is available
    if verbose:
        # verbose case, update progressbar
        console.log("running solver")
        if "FDTD" in task_type:
            with Progress(console=console) as progress:
                pbar_pd = progress.add_task("% done", total=100)
                perc_done, _ = get_run_info(task_id)

                while (
                    perc_done is not None and perc_done < 100 and get_status(task_id) == "running"
                ):
                    perc_done, field_decay = get_run_info(task_id)
                    new_description = f"solver progress (field decay = {field_decay:.2e})"
                    progress.update(pbar_pd, completed=perc_done, description=new_description)
                    time.sleep(RUN_REFRESH_TIME)

                perc_done, field_decay = get_run_info(task_id)
                if perc_done is not None and perc_done < 100 and field_decay > 0:
                    console.log(f"early shutoff detected at {perc_done:1.0f}%, exiting.")

                new_description = f"solver progress (field decay = {field_decay:.2e})"
                progress.update(pbar_pd, completed=100, refresh=True, description=new_description)
        elif task_type == "EME":
            with Progress(console=console) as progress:
                pbar_pd = progress.add_task("% done", total=100)
                perc_done, _ = get_run_info(task_id)

                while (
                    perc_done is not None and perc_done < 100 and get_status(task_id) == "running"
                ):
                    perc_done, _ = get_run_info(task_id)
                    new_description = "solver progress"
                    progress.update(pbar_pd, completed=perc_done, description=new_description)
                    time.sleep(RUN_REFRESH_TIME)

                perc_done, _ = get_run_info(task_id)
                new_description = "solver progress"
                progress.update(pbar_pd, completed=100, refresh=True, description=new_description)
        else:
            while get_status(task_id) == "running":
                perc_done, _ = get_run_info(task_id)
                time.sleep(RUN_REFRESH_TIME)

    else:
        # non-verbose case, just keep checking until status is not running or perc_done >= 100
        perc_done, _ = get_run_info(task_id)
        while perc_done is not None and perc_done < 100 and get_status(task_id) == "running":
            perc_done, field_decay = get_run_info(task_id)
            time.sleep(RUN_REFRESH_TIME)

    # post processing
    if verbose:
        status = get_status(task_id)
        if status != "running":
            console.log(f"status = {status}")

        with console.status(f"[bold green]Finishing '{task_name}'...", spinner="runner"):
            while status not in END_STATES:
                new_status = get_status(task_id)
                if new_status != status:
                    status = new_status
                    console.log(f"status = {status}")
                time.sleep(REFRESH_TIME)

        if task_type in GUI_SUPPORTED_TASK_TYPES:
            url = _get_url(task_id)
            console.log(f"View simulation result at [blue underline][link={url}]'{url}'[/link].")
    else:
        while get_status(task_id) not in END_STATES:
            time.sleep(REFRESH_TIME)


@wait_for_connection
def abort(task_id: TaskId) -> Optional[TaskInfo]:
    """Abort server-side data associated with task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.

    Returns
    -------
    TaskInfo
        Object containing information about status, size, credits of task.
    """
    console = get_logging_console()

    task = TaskFactory.get(task_id, verbose=False)
    if not task:
        return None
    url = task.get_url()
    task.abort()
    console.log(
        f"Task is aborting. View task using web UI at [link={url}]'{url}'[/link] to check the result."
    )
    return TaskInfo(
        **{"taskId": task_id, "taskType": getattr(task, "task_type", None), **task.dict()}
    )


@wait_for_connection
def download(
    task_id: TaskId,
    path: PathLike = "simulation_data.hdf5",
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
) -> None:
    """Download results of task to file.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : PathLike = "simulation_data.hdf5"
        Download path to .hdf5 data file (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.

    """
    path = Path(path)
    task = TaskFactory.get(task_id, verbose=False)
    if isinstance(task, BatchTask):
        if path.name == "simulation_data.hdf5":
            path = path.with_name("cm_data.hdf5")
        task.get_data_hdf5(
            to_file=path,
            remote_data_file_gz=CM_DATA_HDF5_GZ,
            verbose=verbose,
            progress_callback=progress_callback,
        )
        return
    info = get_info(task_id, verbose=False)
    remote_data_file = SIMULATION_DATA_HDF5_GZ
    if info.taskType == "MODE_SOLVER":
        remote_data_file = MODE_DATA_HDF5_GZ
    task.get_data_hdf5(
        to_file=path,
        remote_data_file_gz=remote_data_file,
        verbose=verbose,
        progress_callback=progress_callback,
    )


@wait_for_connection
def download_json(task_id: TaskId, path: PathLike = SIM_FILE_JSON, verbose: bool = True) -> None:
    """Download the ``.json`` file associated with the :class:`.Simulation` of a given task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : PathLike = "simulation.json"
        Download path to .json file of simulation (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.

    """
    task = TaskFactory.get(task_id, verbose=False)
    if isinstance(task, BatchTask):
        raise NotImplementedError("Operation not implemented for modeler batches.")
    task.get_simulation_json(path, verbose=verbose)


@wait_for_connection
def delete_old(days_old: int, folder_name: str = "default") -> int:
    """Remove folder contents older than ``days_old``."""
    folder = Folder.get(folder_name, create=True)
    return folder.delete_old(days_old)


@wait_for_connection
def load_simulation(
    task_id: TaskId, path: PathLike = SIM_FILE_JSON, verbose: bool = True
) -> WorkflowType:
    """Download the ``.json`` file of a task and load the associated simulation.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : PathLike = "simulation.json"
        Download path to .json or .hdf5 file of simulation (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.

    Returns
    -------
    Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`]
        Simulation loaded from downloaded json file.
    """
    task = TaskFactory.get(task_id, verbose=False)
    if isinstance(task, BatchTask):
        raise NotImplementedError("Operation not implemented for modeler batches.")
    path = Path(path)
    if path.suffix == ".json":
        task.get_simulation_json(path, verbose=verbose)
    elif path.suffix == ".hdf5":
        task.get_simulation_hdf5(path, verbose=verbose)
    else:
        raise ValueError("Path suffix must be '.json' or '.hdf5'")
    return Tidy3dStub.from_file(path)


@wait_for_connection
def download_log(
    task_id: TaskId,
    path: PathLike = "tidy3d.log",
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
) -> None:
    """Download the tidy3d log file associated with a task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : PathLike = "tidy3d.log"
        Download path to log file (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.

    Note
    ----
    To load downloaded results into data, call :meth:`load` with option ``replace_existing=False``.
    """
    task = TaskFactory.get(task_id, verbose=False)
    if isinstance(task, BatchTask):
        raise NotImplementedError("Operation not implemented for modeler batches.")
    task.get_log(path, verbose=verbose, progress_callback=progress_callback)


@wait_for_connection
def load(
    task_id: Optional[TaskId],
    path: PathLike = "simulation_data.hdf5",
    replace_existing: bool = True,
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
    lazy: bool = False,
) -> WorkflowDataType:
    """
    Download and Load simulation results into :class:`.SimulationData` object.

    Notes
    -----

        After the simulation is complete, you can load the results into a :class:`.SimulationData` object by its
        ``task_id`` using:

        .. code-block:: python

            sim_data = web.load(task_id, path="outt/sim.hdf5", verbose=verbose)

        The :meth:`tidy3d.web.api.webapi.load` method is very convenient to load and postprocess results from simulations
        created using Tidy3D GUI.

    Parameters
    ----------
    task_id : Optional[str] = None
        Unique identifier of task on server. Returned by :meth:`upload`. If None, file is assumed to exist already from cache.
    path : PathLike
        Download path to .hdf5 data file (including filename).
    replace_existing : bool = True
        Downloads the data even if path exists (overwriting the existing).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.
    lazy : bool = False
        Whether to load the actual data (``lazy=False``) or return a proxy that loads
        the data when accessed (``lazy=True``).

    Returns
    -------
    Union[:class:`.SimulationData`, :class:`.HeatSimulationData`, :class:`.EMESimulationData`]
        Object containing simulation data.
    """
    path = Path(path)
    task = TaskFactory.get(task_id) if task_id else None
    # For component modeler batches, default to a clearer filename if the default was used.
    if (
        task_id
        and isinstance(task, BatchTask)
        and path.name in {"simulation_data.hdf5", "simulation_data.hdf5.gz"}
    ):
        path = path.with_name(path.name.replace("simulation", "cm"))

    if task_id is None:
        if not path.exists():
            raise FileNotFoundError("Cached file not found.")
    elif not path.exists() or replace_existing:
        download(task_id=task_id, path=path, verbose=verbose, progress_callback=progress_callback)

    if verbose and task_id is not None:
        console = get_logging_console()
        if isinstance(task, BatchTask):
            console.log(f"Loading component modeler data from {path}")
        else:
            console.log(f"Loading simulation from {path}")

    stub_data = Tidy3dStubData.postprocess(path, lazy=lazy)

    simulation_cache = resolve_local_cache()
    if simulation_cache is not None and task_id is not None:
        info = get_info(task_id, verbose=False)
        workflow_type = getattr(info, "taskType", None)
        if (
            workflow_type != TaskType.MODE_SOLVER.name
        ):  # we cannot get the simulation from data or web for mode solver
            simulation = None
            if lazy:  # get simulation via web to avoid unpacking of lazy object in store_result
                try:
                    with tempfile.NamedTemporaryFile(suffix=".hdf5") as tmp_file:
                        simulation = load_simulation(task_id, path=tmp_file.name, verbose=False)
                except Exception as e:
                    log.info(f"Failed to load simulation for storing results: {e}.")
                    return stub_data
            simulation_cache.store_result(
                stub_data=stub_data,
                task_id=task_id,
                path=path,
                workflow_type=workflow_type,
                simulation=simulation,
            )

    return stub_data


def _monitor_modeler_batch(
    task_id: str,
    verbose: bool = True,
    max_detail_tasks: int = 20,
) -> None:
    """Monitor modeler batch progress with aggregate and per-task views."""
    console = get_logging_console() if verbose else None
    task = BatchTask.get(task_id=task_id)
    detail = _get_batch_detail_handle_error_status(task)
    name = detail.name or "modeler_batch"
    group_id = detail.groupId
    status = detail.status.lower()

    # Non-verbose path: poll without progress bars then return
    if not verbose:
        # Run phase
        while status_to_stage(status)[0] not in END_STATES:
            time.sleep(REFRESH_TIME)
            detail = _get_batch_detail_handle_error_status(task)
            status = detail.status.lower()

        return

    progress_columns = (
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=25),
        TaskProgressColumn(),
        TextColumn("[progress.description]{task.fields[status]}"),
        TimeElapsedColumn(),
    )
    # Make the header
    header = f"Subtasks status - {name}"
    if group_id:
        header += f"\nGroup ID: '{group_id}'"
    console.log(header)
    with Progress(*progress_columns, console=console, transient=False) as progress:
        # Phase: Run (aggregate + per-task)
        stage = status_to_stage(status)[0]
        p_run = progress.add_task("Run Total", total=1.0, status=f" {stage} ")
        task_bars: dict[str, int] = {}
        prev_stage = status_to_stage(status)[0]
        console.log(f"Batch status = {status}")

        # Note: get_status errors if an erroring status occurred
        end_monitor = False
        while not end_monitor:
            total = len(detail.tasks)
            r = detail.runSuccess or 0
            if stage != prev_stage:
                prev_stage = stage
                console.log(f"Batch status = {stage}")
                progress.update(p_run, status=f" {stage} ")

            # Create per-task bars as soon as tasks appear
            if total and total <= max_detail_tasks and detail.tasks:
                name_to_task = {(t.taskName or t.taskId): t for t in (detail.tasks or [])}
                for name, t in name_to_task.items():
                    if name not in task_bars:
                        tstatus = (t.status or "draft").lower()
                        _, idx = status_to_stage(tstatus)
                        pbar = progress.add_task(
                            f"  {name}",
                            total=1.0,
                            completed=STATE_PROGRESS_PERCENTAGE[tstatus] / 100,
                            status=f" {tstatus} ",
                        )
                        task_bars[name] = pbar

            # Aggregate run progress: average stage fraction across tasks (80% weight)
            # Final 20% achieved only when batch status is completed
            if detail.tasks:
                acc = 0.0
                n_members = 0
                for t in detail.tasks or []:
                    n_members += 1
                    tstatus = (t.status or "draft").lower()
                    _, idx = status_to_stage(tstatus)
                    acc += max(0.0, min(1.0, idx / MAX_STEPS))
                task_avg = (acc / float(n_members)) if n_members else 0.0
                run_frac = task_avg * 0.8
            else:
                run_frac = (r / total) * 0.8 if total else 0.0

            # Final 20% only when batch is completed
            if status in END_STATES:
                # Makes sure last state is logged
                end_monitor = True
                run_frac = 1.0

            progress.update(p_run, completed=run_frac)

            # Update per-task bars
            if task_bars and detail.tasks:
                name_to_task = {(t.taskName or t.taskId): t for t in (detail.tasks or [])}
                for tname, pbar in task_bars.items():
                    t = name_to_task.get(tname)
                    if not t:
                        continue
                    tstatus = (t.status or "draft").lower()
                    _, idx = status_to_stage(tstatus)
                    progress.update(
                        pbar,
                        completed=STATE_PROGRESS_PERCENTAGE[tstatus] / 100,
                        description=f"  {tname}",
                        status=f" {tstatus} ",
                        refresh=False,
                    )

            progress.refresh()
            time.sleep(REFRESH_TIME)
            detail = _get_batch_detail_handle_error_status(task)
            status = detail.status.lower()
            stage = status_to_stage(status)[0]

        if console is not None:
            console.log("Modeler has finished running successfully.")
            real_cost(task.task_id, verbose=verbose)


@wait_for_connection
def delete(task_id: TaskId, versions: bool = False) -> TaskInfo:
    """Delete server-side data associated with task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    versions : bool = False
        If ``True``, delete all versions of the task in the task group. Otherwise, delete only the version associated with the current task ID.

    Returns
    -------
    :class:`.TaskInfo`
        Object containing information about status, size, credits of task.

    """
    if not task_id:
        raise ValueError("Task id not found.")
    task = TaskFactory.get(task_id, verbose=False)
    task.delete(versions)
    return TaskInfo(**{"taskId": task.task_id, **task.dict()})


@wait_for_connection
def download_simulation(
    task_id: TaskId,
    path: PathLike = SIM_FILE_HDF5,
    verbose: bool = True,
    progress_callback: Optional[Callable[[float], None]] = None,
) -> None:
    """Download the ``.hdf5`` file associated with the :class:`.Simulation` of a given task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    path : PathLike = "simulation.hdf5"
        Download path to .hdf5 file of simulation (including filename).
    verbose : bool = True
        If ``True``, will print progressbars and status, otherwise, will run silently.
    progress_callback : Callable[[float], None] = None
        Optional callback function called when downloading file with ``bytes_in_chunk`` as argument.

    """
    task = TaskFactory.get(task_id, verbose=False)
    if isinstance(task, BatchTask):
        raise NotImplementedError("Operation not implemented for modeler batches.")
    info = get_info(task_id, verbose=False)
    remote_sim_file = SIM_FILE_HDF5_GZ
    if info.taskType == "MODE_SOLVER":
        remote_sim_file = MODE_FILE_HDF5_GZ
    task.get_simulation_hdf5(
        path,
        verbose=verbose,
        progress_callback=progress_callback,
        remote_sim_file=remote_sim_file,
    )


@wait_for_connection
def get_tasks(
    num_tasks: Optional[int] = None, order: Literal["new", "old"] = "new", folder: str = "default"
) -> list[dict]:
    """Get a list with the metadata of the last ``num_tasks`` tasks.

    Parameters
    ----------
    num_tasks : int = None
        The number of tasks to return, or, if ``None``, return all.
    order : Literal["new", "old"] = "new"
        Return the tasks in order of newest-first or oldest-first.
    folder: str = "default"
        Folder from which to get the tasks.

    Returns
    -------
    List[Dict]
        List of dictionaries storing the information for each of the tasks last ``num_tasks`` tasks.
    """
    folder = Folder.get(folder, create=True)
    tasks = folder.list_tasks()
    if not tasks:
        return []
    if order == "new":
        tasks = sorted(tasks, key=lambda t: t.created_at, reverse=True)
    elif order == "old":
        tasks = sorted(tasks, key=lambda t: t.created_at)
    if num_tasks is not None:
        tasks = tasks[:num_tasks]
    return [task.dict() for task in tasks]


@wait_for_connection
def estimate_cost(
    task_id: str, verbose: bool = True, solver_version: Optional[str] = None
) -> float:
    """Compute the maximum FlexCredit charge for a given task.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    verbose : bool = True
        Whether to log the cost and helpful messages.
    solver_version : str = None
        Target solver version.

    Returns
    -------
    float
        Estimated maximum cost for :class:`.Simulation` associated with given ``task_id``.

    Note
    ----
    Cost is calculated assuming the simulation runs for
    the full ``run_time``. If early shut-off is triggered, the cost is adjusted proportionately.
    A minimum simulation cost may also apply, which depends on the task details.

    Notes
    -----

        We can get the cost estimate of running the task before actually running it. This prevents us from
        accidentally running large jobs that we set up by mistake. The estimated cost is the maximum cost
        corresponding to running all the time steps.

    Examples
    --------

    Basic example:

    .. code-block:: python

        # initializes job, puts task on server (but doesn't run it)
        job = web.Job(simulation=sim, task_name="job", verbose=verbose)

        # estimate the maximum cost
        estimated_cost = web.estimate_cost(job.task_id)

        print(f'The estimated maximum cost is {estimated_cost:.3f} Flex Credits.')

    """
    if not isinstance(task_id, str):
        raise ValueError(
            f"Task ID: {task_id} is not a string. You can get it using 'web.upload(<WorkflowType>)'."
        )

    console = get_logging_console() if verbose else None

    task = TaskFactory.get(task_id, verbose=False)
    detail = task.detail()
    if isinstance(task, BatchTask):
        check_task_type = "FDTD" if detail.taskType == "MODAL_CM" else "RF_FDTD"
        task.check(solver_version=solver_version, check_task_type=check_task_type)
        detail = task.detail()
        status = detail.status.lower()
        while status not in ALL_POST_VALIDATE_STATES:
            time.sleep(REFRESH_TIME)
            detail = task.detail()
            status = detail.status.lower()
        if status in ERROR_STATES:
            _batch_detail_error(resource_id=task_id)
        est_flex_unit = detail.estFlexUnit
        if verbose:
            console.log(
                f"Maximum FlexCredit cost: {est_flex_unit:1.3f}. Minimum cost depends on "
                "task execution details. Use 'web.real_cost(task_id)' after run."
            )
        return est_flex_unit

    # simulation path
    task.estimate_cost(solver_version=solver_version)
    task_info = get_info(task_id)
    status = task_info.metadataStatus

    # Wait for a termination status
    while status not in ALL_POST_VALIDATE_STATES:
        time.sleep(REFRESH_TIME)
        task_info = get_info(task_id)
        status = task_info.metadataStatus

    if status in ERROR_STATES:
        try:
            # Try to obtain the error message
            task = SimulationTask(taskId=task_id)
            with tempfile.NamedTemporaryFile(suffix=".json") as tmp_file:
                task.get_error_json(to_file=tmp_file.name, validation=True)
                with open(tmp_file.name) as f:
                    error_content = json.load(f)
                    error_msg = error_content["validation_error"]
        except Exception:
            # If the error message could not be obtained, raise a generic error message
            error_msg = "Error message could not be obtained, please contact customer support."
        raise WebError(f"Error estimating cost for task {task_id}! {error_msg}")
    if verbose:
        console.log(
            f"Estimated FlexCredit cost: {task_info.estFlexUnit:1.3f}. Minimum cost depends on "
            "task execution details. Use 'web.real_cost(task_id)' to get the billed FlexCredit "
            "cost after a simulation run."
        )
        fc_mode = task_info.estFlexCreditMode
        fc_post = task_info.estFlexCreditPostProcess
        if fc_mode:
            console.log(f"  {fc_mode:1.3f} FlexCredit of the total cost from mode solves.")
        if fc_post:
            console.log(f"  {fc_post:1.3f} FlexCredit of the total cost from post-processing.")
    return task_info.estFlexUnit


@wait_for_connection
def real_cost(task_id: str, verbose: bool = True) -> float | None:
    """Get the billed cost for given task after it has been run.

    Parameters
    ----------
    task_id : str
        Unique identifier of task on server.  Returned by :meth:`upload`.
    verbose : bool = True
        Whether to log the cost and helpful messages.

    Returns
    -------
    float
        The flex credit cost that was billed for the given ``task_id``.

    Note
    ----
        The billed cost may not be immediately available when the task status is set to ``success``,
        but should be available shortly after.

    Examples
    --------

    To obtain the cost of a simulation, you can use the function ``tidy3d.web.real_cost(task_id)``. In the example
    below, a job is created, and its cost is estimated. After running the simulation, the real cost can be obtained.

    .. code-block:: python

        import time

        # initializes job, puts task on server (but doesn't run it)
        job = web.Job(simulation=sim, task_name="job", verbose=verbose)

        # estimate the maximum cost
        estimated_cost = web.estimate_cost(job.task_id)

        print(f'The estimated maximum cost is {estimated_cost:.3f} Flex Credits.')

        # Runs the simulation.
        sim_data = job.run(path="data/sim_data.hdf5")

        time.sleep(5)

        # Get the billed FlexCredit cost after a simulation run.
        cost = web.real_cost(job.task_id)
    """
    if not isinstance(task_id, str):
        raise ValueError(
            f"Task ID: {task_id} is not a string. You can get it using 'web.upload(<WorkflowType>)'."
        )

    console = get_logging_console() if verbose else None
    task_info = get_info(task_id)
    flex_unit = task_info.realFlexUnit
    ori_flex_unit = getattr(task_info, "oriRealFlexUnit", flex_unit)
    if not flex_unit:
        log.warning(
            f"Billed FlexCredit for task '{task_id}' is not available. If the task has been "
            "successfully run, it should be available shortly."
        )
    else:
        if verbose:
            console.log(f"Billed flex credit cost: {flex_unit:1.3f}.")
            if flex_unit != ori_flex_unit and "FDTD" in task_info.taskType:
                console.log(
                    "Note: the task cost pro-rated due to early shutoff was below the minimum "
                    "threshold, due to fast shutoff. Decreasing the simulation 'run_time' should "
                    "decrease the estimated, and correspondingly the billed cost of such tasks."
                )
    return flex_unit


@wait_for_connection
def account(verbose: bool = True) -> Account:
    """Get account information including FlexCredit balance and usage limits.

    Parameters
    ----------
    verbose : bool = True
        If ``True``, prints account information including credit balance, expiration,
        and free simulation counts.

    Returns
    -------
    Account
        Object containing account information such as credit balance, expiration dates,
        and daily free simulation counts.

    Examples
    --------
    Get account information:

    .. code-block:: python

        account_info = web.account()
        # Displays:
        # Current FlexCredit balance: 10.00 and expiration date: 2024-12-31 23:59:59.
        # Remaining daily free simulations: 3.
    """
    account_info = Account.get()
    if verbose and account_info:
        console = get_logging_console()
        credit = account_info.credit
        credit_expiration = account_info.credit_expiration
        cycle_type = account_info.allowance_cycle_type
        cycle_amount = account_info.allowance_current_cycle_amount
        cycle_end_date = account_info.allowance_current_cycle_end_date
        free_simulation_counts = account_info.daily_free_simulation_counts

        message = ""
        if credit is not None:
            message += f"Current FlexCredit balance: {credit:.2f}"
            if credit_expiration is not None:
                message += (
                    f" and expiration date: {credit_expiration.strftime('%Y-%m-%d %H:%M:%S')}. "
                )
            else:
                message += ". "
        if cycle_type is not None and cycle_amount is not None and cycle_end_date is not None:
            cycle_end = cycle_end_date.strftime("%Y-%m-%d %H:%M:%S")
            message += f"{cycle_type} FlexCredit balance: {cycle_amount:.2f} and expiration date: {cycle_end}. "
        if free_simulation_counts is not None:
            message += f"Remaining daily free simulations: {free_simulation_counts}."

        console.log(message)

    return account_info


@wait_for_connection
def test() -> None:
    """Confirm whether Tidy3D authentication is configured.

    Raises
    ------
    WebError
        If Tidy3D authentication is not configured correctly.

    Notes
    -----
    This method tests the authentication configuration by attempting to retrieve
    the task list. If authentication is not properly set up, it will raise an
    exception with instructions on how to configure authentication.

    Examples
    --------
    Test authentication:

    .. code-block:: python

        web.test()
        # If successful, displays:
        # Authentication configured successfully!
    """
    try:
        # note, this is a little slow, but the only call that doesn't require providing a task id.
        get_tasks(num_tasks=0)
        console = get_logging_console()
        console.log("Authentication configured successfully!")
    except (WebError, HTTPError) as e:
        url = "https://docs.flexcompute.com/projects/tidy3d/en/latest/index.html"
        msg = (
            str(e)
            + "\n\n"
            + "It looks like the Tidy3D Python interface is not configured with your "
            "unique API key. "
            "To get your API key, sign into 'https://tidy3d.simulation.cloud' and copy it "
            "from your 'Account' page. Then you can configure tidy3d through command line "
            "'tidy3d configure' (recommended). Alternatively, one can manually create the configuration "
            "file by creating a file at your home directory '~/.tidy3d/config' (unix) or "
            "'.tidy3d/config' (windows) with content like: \n\n"
            "apikey = 'XXX' \n\nHere XXX is your API key copied from your account page within quotes.\n\n"
            f"For details, check the instructions at {url}."
        )
        raise WebError(msg) from e



from __future__ import annotations

from collections.abc import Generator

import pytest
from tidy3d_frontend.tidy3d.web.core.task_core import TaskFactory


@pytest.fixture(autouse=True)
def clear_task_factory_registry() -> Generator[None, None, None]:
    """Ensure TaskFactory registry is empty for each test."""
    TaskFactory.reset()
    TaskFactory.reset()
    yield
    TaskFactory.reset()



"""Constants for the CLI."""

from __future__ import annotations

from tidy3d.config.loader import resolve_config_directory

_CONFIG_ROOT = resolve_config_directory()

TIDY3D_DIR = str(_CONFIG_ROOT)
CONFIG_FILE = str(_CONFIG_ROOT / "config.toml")



from __future__ import annotations

from typing import Optional

import numpy as np

from tidy3d.components.microwave.data.monitor_data import AntennaMetricsData
from tidy3d.plugins.smatrix.data.data_array import PortDataArray
from tidy3d.plugins.smatrix.data.terminal import TerminalComponentModelerData
from tidy3d.plugins.smatrix.types import NetworkIndex


def get_antenna_metrics_data(
    terminal_component_modeler_data: TerminalComponentModelerData,
    port_amplitudes: Optional[dict[NetworkIndex, complex]] = None,
    monitor_name: Optional[str] = None,
) -> AntennaMetricsData:
    """Calculate antenna parameters using superposition of fields from multiple port excitations.

    The method computes the radiated far fields and port excitation power wave amplitudes
    for a superposition of port excitations, which can be used to analyze antenna radiation
    characteristics.

    Note
    ----
    The ``NetworkIndex`` identifies a single excitation in the modeled device, so it represents
    a :class:`.LumpedPort` or a single mode from a :class:`.WavePort`. Use the static method
    :meth:`.TerminalComponentModeler.network_index` to convert port and optional mode index
    into the appropriate ``NetworkIndex`` for use in the ``port_amplitudes`` dictionary.

    Parameters
    ----------
    terminal_component_modeler_data: TerminalComponentModelerData
        Data associated with a :class:`.TerminalComponentModeler` simulation run.
    port_amplitudes : dict[NetworkIndex, complex] = None
        Dictionary mapping a network index to their desired excitation amplitudes. For each network port,
        :math:`\\frac{1}{2}|a|^2` represents the incident power from that port into the system.
        If ``None``, uses only the first port without any scaling of the raw simulation data. When
        ``None`` is passed as a port amplitude, the raw simulation data is used for that port. Note
        that in this method ``a`` represents the incident wave amplitude using the power wave definition
        in [2].
    monitor_name : str = None
        Name of the :class:`.DirectivityMonitor` to use for calculating far fields.
        If ``None``, uses the first monitor in ``radiation_monitors``.

    Returns
    -------
    :class:`.AntennaMetricsData`
        Container with antenna parameters including directivity, gain, and radiation efficiency,
        computed from the superposition of fields from all excited ports.
    """
    # Use the first port as default if none specified
    if port_amplitudes is None:
        first_port_index = terminal_component_modeler_data.modeler.matrix_indices_source[0]
        port_amplitudes = {first_port_index: None}
    # Get the radiation monitor, use first as default
    # if none specified
    if monitor_name is None:
        rad_mon = terminal_component_modeler_data.modeler._finalized_radiation_monitors[0]
    else:
        rad_mon = terminal_component_modeler_data.modeler.get_radiation_monitor_by_name(
            monitor_name
        )

    # Create data arrays for holding the superposition of all port power wave amplitudes
    f = list(rad_mon.freqs)
    coords = {"f": f, "port": list(terminal_component_modeler_data.modeler.matrix_indices_monitor)}
    a_sum = PortDataArray(
        np.zeros(
            (len(f), len(terminal_component_modeler_data.modeler.matrix_indices_monitor)),
            dtype=complex,
        ),
        coords=coords,
    )
    b_sum = a_sum.copy()
    a_matrix, b_matrix = terminal_component_modeler_data.port_power_wave_matrices
    # Retrieve associated simulation data
    combined_directivity_data = None
    for port_in_index, amplitude in port_amplitudes.items():
        port, mode_index = terminal_component_modeler_data.modeler.network_dict[port_in_index]
        if amplitude is not None:
            if np.isclose(amplitude, 0.0):
                continue
        sim_data_port = terminal_component_modeler_data.data[
            terminal_component_modeler_data.modeler.get_task_name(port, mode_index)
        ]

        a, b = (
            a_matrix.sel(port_in=port_in_index, drop=True),
            b_matrix.sel(port_in=port_in_index, drop=True),
        )

        # Select a possible subset of frequencies
        a = a.sel(f=f)
        b = b.sel(f=f)
        a_raw = a.sel(port_out=port_in_index)

        if amplitude is None:
            # No scaling performed when amplitude is None
            scaled_directivity_data = sim_data_port[rad_mon.name]
            scale_factor = 1.0
        else:
            scaled_directivity_data = (
                terminal_component_modeler_data._monitor_data_at_port_amplitude(
                    port_in_index, rad_mon.name, amplitude, a_raw
                )
            )
            scale_factor = amplitude / a_raw
        a = scale_factor * a
        b = scale_factor * b

        # Combine the possibly scaled directivity data and the power wave amplitudes
        if combined_directivity_data is None:
            combined_directivity_data = scaled_directivity_data
        else:
            combined_directivity_data = combined_directivity_data + scaled_directivity_data
        a_sum += a.rename({"port_out": "port"})
        b_sum += b.rename({"port_out": "port"})

    # Compute and add power measures to results
    power_incident = np.real(0.5 * a_sum * np.conj(a_sum)).sum(dim="port")
    power_reflected = np.real(0.5 * b_sum * np.conj(b_sum)).sum(dim="port")
    return AntennaMetricsData.from_directivity_data(
        combined_directivity_data, power_incident, power_reflected
    )



"""Cache-related CLI commands."""

from __future__ import annotations

from typing import Optional

import click

from tidy3d import config
from tidy3d.web.cache import LocalCache, resolve_local_cache
from tidy3d.web.cache import clear as clear_cache


def _fmt_size(num_bytes: int) -> str:
    """Format bytes into human-readable form."""
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if num_bytes < 1024.0 or unit == "TB":
            return f"{num_bytes:.2f} {unit}"
        num_bytes /= 1024.0
    return f"{num_bytes:.2f} B"  # fallback, though unreachable


def _get_cache(ensure: bool = True) -> Optional[LocalCache]:
    """Resolve the local cache object, surfacing errors as ClickExceptions."""
    try:
        cache = resolve_local_cache(use_cache=True)
    except Exception as exc:  # pragma: no cover - defensive guard
        raise click.ClickException(f"Failed to access local cache: {exc}") from exc
    if cache is None and ensure:
        raise click.ClickException("Local cache is disabled in the current configuration.")
    return cache


@click.group(name="cache")
def cache_group() -> None:
    """Inspect or manage the local cache."""


@cache_group.command()
def info() -> None:
    """Display current cache configuration and usage statistics."""

    enabled = bool(config.local_cache.enabled)
    directory = config.local_cache.directory
    max_entries = config.local_cache.max_entries
    max_size_gb = config.local_cache.max_size_gb

    cache = _get_cache(ensure=False)
    entries = 0
    total_size = 0
    if cache is not None:
        stats = cache.sync_stats()
        entries = stats.total_entries
        total_size = stats.total_size

    click.echo(f"Enabled: {'yes' if enabled else 'no'}")
    click.echo(f"Directory: {directory}")
    click.echo(f"Entries: {entries}")
    click.echo(f"Total size: {_fmt_size(total_size)}")
    click.echo("Max entries: " + (str(max_entries) if max_entries else "unlimited"))
    click.echo(
        "Max size: " + (f"{_fmt_size(max_size_gb * 1024**3)}" if max_size_gb else "unlimited")
    )


@cache_group.command(name="list")
def list() -> None:
    """List cached entries in a readable, separated format."""

    cache = _get_cache()
    entries = cache.list()
    if not entries:
        click.echo("Cache is empty.")
        return

    def fmt_key(key: str) -> str:
        return key.replace("_", " ").capitalize()

    for i, entry in enumerate(entries, start=1):
        entry.pop("simulation_hash", None)
        entry.pop("checksum", None)
        entry["file_size"] = _fmt_size(entry["file_size"])

        click.echo(f"\n=== Cache Entry #{i} ===")
        for k, v in entry.items():
            click.echo(f"{fmt_key(k)}: {v}")
    click.echo("")


@cache_group.command()
def clear() -> None:
    """Remove all cache contents."""

    clear_cache()
    click.echo("Local cache cleared.")



from __future__ import annotations

# progression order for a typical run
PROGRESSION_ORDER = (
    "draft",
    "queued",
    "preprocess",
    "running",
    "postprocess",
    "success",
)

MAX_STEPS = 4
COMPLETED_PERCENT = 100

PRE_ERROR_STATES = {
    "aborting",
}

ERROR_STATES = {
    "validate_error",
    "error",
    "errored",
    "blocked",
    "preprocess_error",
    "run_error",
    "aborted",
    "deleted",
    "postprocess_error",
}

PRE_VALIDATE_STATES = {
    "draft",
    "validating",
    "validate",
}

QUEUED_STATES = {"queued", "queued_solver"}

PREPROCESS_STATES = {"preprocess"}

RUNNING_STATES = {"running", "preprocess_success"}

POSTPROCESS_STATES = {
    "postprocess",
    "run_success",
}

SUCCESS_STATES = {
    "visualize",
    "success",
    "completed",
    "processed",
    "postprocess_success",
}

DIVERGED_STATES = {
    "diverge",
    "diverged",
}

COMPLETED_STATES = DIVERGED_STATES | SUCCESS_STATES

END_STATES = ERROR_STATES | COMPLETED_STATES

POST_VALIDATE_STATES = {"validate_success", "validate_warn", "warning"}

DRAFT_STATES = PRE_VALIDATE_STATES | POST_VALIDATE_STATES

ALL_POST_VALIDATE_STATES = POST_VALIDATE_STATES | RUNNING_STATES | POSTPROCESS_STATES | END_STATES

VALID_PROGRESS_STATES = (
    DRAFT_STATES
    | QUEUED_STATES
    | POST_VALIDATE_STATES
    | RUNNING_STATES
    | POSTPROCESS_STATES
    | COMPLETED_STATES
    | PRE_ERROR_STATES
)

ALL_STATES = VALID_PROGRESS_STATES | ERROR_STATES

STATE_PROGRESS_PERCENTAGE = dict.fromkeys(ALL_STATES, 0)
STATE_PROGRESS_PERCENTAGE.update(dict.fromkeys(COMPLETED_STATES, COMPLETED_PERCENT))
STATE_PROGRESS_PERCENTAGE.update(
    {state: round((0 / MAX_STEPS) * COMPLETED_PERCENT) for state in QUEUED_STATES}
)
STATE_PROGRESS_PERCENTAGE.update(
    {state: round((0 / MAX_STEPS) * COMPLETED_PERCENT) for state in DIVERGED_STATES}
)
STATE_PROGRESS_PERCENTAGE.update(
    {state: round((1 / MAX_STEPS) * COMPLETED_PERCENT) for state in PREPROCESS_STATES}
)
STATE_PROGRESS_PERCENTAGE.update(
    {state: round((2 / MAX_STEPS) * COMPLETED_PERCENT) for state in RUNNING_STATES}
)
STATE_PROGRESS_PERCENTAGE.update(
    {state: round((3 / MAX_STEPS) * COMPLETED_PERCENT) for state in POSTPROCESS_STATES}
)
STATE_PROGRESS_PERCENTAGE.update(
    {state: round((4 / MAX_STEPS) * COMPLETED_PERCENT) for state in SUCCESS_STATES}
)


def status_to_stage(status: str) -> tuple[str, int]:
    """Map task status to monotonic stage for progress bars.

    Parameters
    ----------
    status : str
        The task status string.

    Returns
    -------
    tuple[str, int]
        A tuple of (stage_name, stage_index) where stage_index corresponds
        to the position in PROGRESSION_ORDER.
    """
    s = (status or "").lower()
    if s in DRAFT_STATES:
        return ("draft", 0)
    if s in QUEUED_STATES:
        return ("queued", 0)
    if s in PREPROCESS_STATES:
        return ("preprocess", 1)
    if s in RUNNING_STATES:
        return ("running", 2)
    if s in POSTPROCESS_STATES:
        return ("postprocess", 3)
    if s in COMPLETED_STATES:
        return ("success", 4)
    # Unknown states map to earliest stage to avoid showing 100% prematurely
    return (s or "unknown", 0)



from __future__ import annotations

from pathlib import Path
from typing import Any

import tidy3d as td
from tidy3d.web.api.container import DEFAULT_DATA_PATH, Batch, Job

from .io_utils import get_vjp_traced_fields, upload_sim_fields_keys


def parse_run_kwargs(**run_kwargs: Any) -> dict[str, Any]:
    """Parse the ``run_kwargs`` to extract what should be passed to the ``Job``/``Batch`` init."""
    job_fields = [*list(Job._upload_fields), "solver_version", "pay_type", "lazy"]
    job_init_kwargs = {k: v for k, v in run_kwargs.items() if k in job_fields}
    return job_init_kwargs


def _run_tidy3d(
    simulation: td.Simulation, task_name: str, **run_kwargs: Any
) -> tuple[td.SimulationData, str]:
    """Run a simulation without any tracers using regular web.run()."""

    job_init_kwargs = parse_run_kwargs(**run_kwargs)
    job = Job(simulation=simulation, task_name=task_name, **job_init_kwargs)
    td.log.info(f"running {job.simulation_type} simulation with '_run_tidy3d()'")
    if job.simulation_type == "autograd_fwd":
        verbose = run_kwargs.get("verbose", False)
        upload_sim_fields_keys(run_kwargs["sim_fields_keys"], task_id=job.task_id, verbose=verbose)
    path = Path(run_kwargs.get("path", DEFAULT_DATA_PATH))
    priority = run_kwargs.get("priority")
    if task_name.endswith("_adjoint"):
        suffixes = "".join(path.suffixes)
        base_name = path.name
        base_without_suffix = base_name[: -len(suffixes)] if suffixes else base_name
        path = path.with_name(f"{base_without_suffix}_adjoint{suffixes}")
    data = job.run(path, priority=priority)
    return data, job.task_id


def _run_async_tidy3d(
    simulations: dict[str, td.Simulation], **run_kwargs: Any
) -> tuple[td.web.api.container.BatchData, dict[str, str]]:
    """Run a batch of simulations using regular web.run()."""

    batch_init_kwargs = parse_run_kwargs(**run_kwargs)
    path_dir = run_kwargs.pop("path_dir", None)
    priority = run_kwargs.get("priority")
    batch = Batch(simulations=simulations, **batch_init_kwargs)
    td.log.info(f"running {batch.simulation_type} batch with '_run_async_tidy3d()'")

    if batch.simulation_type == "autograd_fwd":
        verbose = run_kwargs.get("verbose", False)
        # Need to upload to get the task_ids
        sims = {
            task_name: sim.updated_copy(simulation_type="autograd_fwd", deep=False)
            for task_name, sim in batch.simulations.items()
        }
        batch = batch.updated_copy(simulations=sims)

        batch.upload()
        task_ids = {key: job.task_id for key, job in batch.jobs.items()}
        for task_name, sim_fields_keys in run_kwargs["sim_fields_keys_dict"].items():
            task_id = task_ids[task_name]
            upload_sim_fields_keys(sim_fields_keys, task_id=task_id, verbose=verbose)

    if path_dir is not None:
        batch_data = batch.run(path_dir, priority=priority)
    else:
        batch_data = batch.run(priority=priority)

    task_ids = {key: job.task_id for key, job in batch.jobs.items()}
    return batch_data, task_ids


def _run_async_tidy3d_bwd(
    simulations: dict[str, td.Simulation],
    **run_kwargs: Any,
) -> dict[str, dict]:
    """Run a batch of adjoint simulations using regular web.run()."""

    batch_init_kwargs = parse_run_kwargs(**run_kwargs)
    _ = run_kwargs.pop("path_dir", None)
    batch = Batch(simulations=simulations, **batch_init_kwargs)
    td.log.info(f"running {batch.simulation_type} batch with '_run_async_tidy3d_bwd()'")

    priority = run_kwargs.get("priority")
    batch.start(priority=priority)
    batch.monitor()

    vjp_traced_fields_dict = {}
    for task_name, job in batch.jobs.items():
        task_id = job.task_id
        vjp = get_vjp_traced_fields(task_id_adj=task_id, verbose=batch.verbose)
        vjp_traced_fields_dict[task_name] = vjp

    return vjp_traced_fields_dict



"""Specification for impedance computation in transmission lines and waveguides."""

from __future__ import annotations

from typing import Optional, Union

import pydantic.v1 as pd

from tidy3d.components.base import skip_if_fields_missing
from tidy3d.components.microwave.base import MicrowaveBaseModel
from tidy3d.components.microwave.path_integrals.types import (
    CurrentPathSpecType,
    VoltagePathSpecType,
)
from tidy3d.exceptions import SetupError


class AutoImpedanceSpec(MicrowaveBaseModel):
    """Specification for fully automatic transmission line impedance computation.

    Notes
    -----
        Automatically calculates impedance using paths based on simulation geometry
        and conductors that intersect the mode plane. No user-defined path
        specifications are required.
    """


class CustomImpedanceSpec(MicrowaveBaseModel):
    """Specification for custom transmission line voltages and currents in mode solvers.

    Notes
    -----
        The :class:`.CustomImpedanceSpec` class specifies how quantities related to transmission line
        modes are computed. It defines the paths for line integrals, which are used to
        compute voltage, current, and characteristic impedance of the transmission line.

        Users must supply at least one of voltage or current path specifications to control where these integrals
        are evaluated. Both voltage_spec and current_spec cannot be ``None`` simultaneously.

    Example
    -------
    >>> from tidy3d.components.microwave.path_integrals.specs.voltage import AxisAlignedVoltageIntegralSpec
    >>> from tidy3d.components.microwave.path_integrals.specs.current import AxisAlignedCurrentIntegralSpec
    >>> voltage_spec = AxisAlignedVoltageIntegralSpec(
    ...     center=(0, 0, 0), size=(0, 0, 1), sign="+"
    ... )
    >>> current_spec = AxisAlignedCurrentIntegralSpec(
    ...     center=(0, 0, 0), size=(2, 1, 0), sign="+"
    ... )
    >>> impedance_spec = CustomImpedanceSpec(
    ...     voltage_spec=voltage_spec,
    ...     current_spec=current_spec
    ... )
    """

    voltage_spec: Optional[VoltagePathSpecType] = pd.Field(
        None,
        title="Voltage Integration Path",
        description="Path specification for computing the voltage associated with a mode profile.",
    )

    current_spec: Optional[CurrentPathSpecType] = pd.Field(
        None,
        title="Current Integration Path",
        description="Path specification for computing the current associated with a mode profile.",
    )

    @pd.validator("current_spec", always=True)
    @skip_if_fields_missing(["voltage_spec"])
    def check_path_spec_combinations(cls, val, values):
        """Validate that at least one of voltage_spec or current_spec is provided.

        In order to define voltage/current/impedance, either a voltage or current path specification
        must be provided. Both cannot be ``None`` simultaneously.
        """

        voltage_spec = values["voltage_spec"]
        if val is None and voltage_spec is None:
            raise SetupError(
                "Not a valid 'CustomImpedanceSpec', the 'voltage_spec' and 'current_spec' cannot both be 'None'."
            )
        return val


ImpedanceSpecType = Union[AutoImpedanceSpec, CustomImpedanceSpec]
